# coding: utf-8

import threading
from time import sleep
from queue import Queue

from Quartz import (
    CFMachPortCreateRunLoopSource,
    CFMachPortInvalidate,
    CFRunLoopAddSource,
    CFRunLoopSourceInvalidate,
    CFRunLoopGetCurrent,
    CFRunLoopRun,
    CFRunLoopStop,
    CFRelease,
    CGEventCreateKeyboardEvent,
    CGEventGetFlags,
    CGEventGetIntegerValueField,
    CGEventKeyboardSetUnicodeString,
    CGEventMaskBit,
    CGEventPost,
    CGEventSetFlags,
    CGEventSourceCreate,
    CGEventTapCreate,
    CGEventTapEnable,
    kCFRunLoopCommonModes,
    kCGEventFlagMaskAlternate,
    kCGEventFlagMaskCommand,
    kCGEventFlagMaskControl,
    kCGEventFlagMaskNonCoalesced,
    kCGEventFlagMaskNumericPad,
    kCGEventFlagMaskSecondaryFn,
    kCGEventFlagMaskShift,
    kCGEventKeyDown,
    kCGEventKeyUp,
    kCGEventTapDisabledByTimeout,
    kCGEventTapOptionDefault,
    kCGHeadInsertEventTap,
    kCGKeyboardEventKeycode,
    kCGSessionEventTap,
    kCGEventSourceStateHIDSystemState,
    NSEvent,
    NSSystemDefined,
)

from plover.oslayer.osxkeyboardlayout import KeyboardLayout
from plover.key_combo import add_modifiers_aliases, parse_key_combo, KEYNAME_TO_CHAR
import plover.log


BACK_SPACE = 51

NX_KEY_OFFSET = 65536

NX_KEYS = {
    "AudioRaiseVolume": 0,
    "AudioLowerVolume": 1,
    "MonBrightnessUp": 2,
    "MonBrightnessDown": 3,
    "AudioMute": 7,
    "Num_Lock": 10,
    "Eject": 14,
    "AudioPause": 16,
    "AudioPlay": 16,
    "AudioNext": 17,
    "AudioPrev": 18,
    "AudioRewind": 20,
    "KbdBrightnessUp": 21,
    "KbdBrightnessDown": 22
}

KEYNAME_TO_KEYCODE = {
    'fn': 63,

    'alt_l': 58, 'alt_r': 61, 'control_l': 59, 'control_r': 62,
    'shift_l': 56, 'shift_r': 60, 'super_l': 55, 'super_r': 55,
    'caps_lock': 57,

    'return': 36, 'tab': 48, 'backspace': BACK_SPACE, 'delete': 117,
    'escape': 53, 'clear': 71,

    'up': 126, 'down': 125, 'left': 123, 'right': 124, 'page_up': 116,
    'page_down': 121, 'home': 115, 'end': 119,

    'f1': 122, 'f2': 120, 'f3': 99, 'f4': 118, 'f5': 96, 'f6': 97,
    'f7': 98, 'f8': 100, 'f9': 101, 'f10': 109, 'f11': 103,
    'f12': 111, 'f13': 105, 'f14': 107, 'f15': 113, 'f16': 106,
    'f17': 64, 'f18': 79, 'f19': 80, 'f20': 90,

    'kp_0': 82, 'kp_1': 83, 'kp_2': 84, 'kp_3': 85, 'kp_4': 86,
    'kp_5': 87, 'kp_6': 88, 'kp_7': 89, 'kp_8': 91, 'kp_9': 92,
    'kp_add': 69, 'kp_decimal': 65, 'kp_delete': 71, 'kp_divide': 75,
    'kp_enter': 76, 'kp_equal': 81, 'kp_multiply': 67, 'kp_subtract': 78,
}
for name, code in NX_KEYS.items():
    KEYNAME_TO_KEYCODE[name.lower()] = code + NX_KEY_OFFSET
add_modifiers_aliases(KEYNAME_TO_KEYCODE)

DEADKEY_SYMBOLS = {
    'dead_acute': '´',
    'dead_circumflex': '^',
    'dead_diaeresis': '¨',
    'dead_grave': '`',
    'dead_tilde': '~',
}

def down(seq):
    return [(x, True) for x in seq]


def up(seq):
    return [(x, False) for x in reversed(seq)]


def down_up(seq):
    return down(seq) + up(seq)

# Maps from keycodes to corresponding event masks.
MODIFIER_KEYS_TO_MASKS = {
    58: kCGEventFlagMaskAlternate,
    61: kCGEventFlagMaskAlternate,
    # As of Sierra we *require* secondary fn to get control to work properly.
    59: kCGEventFlagMaskControl,
    62: kCGEventFlagMaskControl,
    56: kCGEventFlagMaskShift,
    60: kCGEventFlagMaskShift,
    55: kCGEventFlagMaskCommand,
    63: kCGEventFlagMaskSecondaryFn,
}

OUTPUT_SOURCE = CGEventSourceCreate(kCGEventSourceStateHIDSystemState)

# For the purposes of this class, we're only watching these keys.
# We could calculate the keys, but our default layout would be misleading:
#
#     KEYCODE_TO_KEY = {keycode: mac_keycode.CharForKeycode(keycode) \
#             for keycode in range(127)}
KEYCODE_TO_KEY = {
    122: 'F1', 120: 'F2', 99: 'F3', 118: 'F4', 96: 'F5', 97: 'F6',
    98: 'F7', 100: 'F8', 101: 'F9', 109: 'F10', 103: 'F11', 111: 'F12',

    50: '`', 29: '0', 18: '1', 19: '2', 20: '3', 21: '4', 23: '5',
    22: '6', 26: '7', 28: '8', 25: '9', 27: '-', 24: '=',

    12: 'q', 13: 'w', 14: 'e', 15: 'r', 17: 't', 16: 'y',
    32: 'u', 34: 'i', 31: 'o',  35: 'p', 33: '[', 30: ']', 42: '\\',

    0: 'a', 1: 's', 2: 'd', 3: 'f', 5: 'g',
    4: 'h', 38: 'j', 40: 'k', 37: 'l', 41: ';', 39: '\'',

    6: 'z', 7: 'x', 8: 'c', 9: 'v', 11: 'b',
    45: 'n', 46: 'm', 43: ',', 47: '.', 44: '/',

    49: 'space', BACK_SPACE: "BackSpace",
    117: "Delete", 125: "Down", 119: "End",
    53: "Escape", 115: "Home", 123: "Left", 121: "Page_Down", 116: "Page_Up",
    36: "Return", 124: "Right", 48: "Tab", 126: "Up",
}

def keycode_needs_fn_mask(keycode):
    return (
        keycode >= KEYNAME_TO_KEYCODE['escape']
        and keycode not in MODIFIER_KEYS_TO_MASKS
    )

class KeyboardCapture(threading.Thread):
    """Implementation of KeyboardCapture for OSX."""

    _KEYBOARD_EVENTS = {kCGEventKeyDown, kCGEventKeyUp}

    def __init__(self):
        threading.Thread.__init__(self, name="KeyboardEventTapThread")
        self._loop = None
        self._event_queue = Queue()  # Drained by event handler thread.

        self._suppressed_keys = set()
        self.key_down = lambda key: None
        self.key_up = lambda key: None

        # Returning the event means that it is passed on
        # for further processing by others.
        #
        # Returning None means that the event is intercepted.
        #
        # Delaying too long in returning appears to cause the
        # system to ignore the tap forever after
        # (https://github.com/openstenoproject/plover/issues/484#issuecomment-214743466).
        #
        # This motivates pushing callbacks to the other side
        # of a queue of received events, so that we can return
        # from this callback as soon as possible.
        def callback(proxy, event_type, event, reference):
            SUPPRESS_EVENT = None
            PASS_EVENT_THROUGH = event

            # Don't pass on meta events meant for this event tap.
            is_unexpected_event = event_type not in self._KEYBOARD_EVENTS
            if is_unexpected_event:
                if event_type == kCGEventTapDisabledByTimeout:
                    # Re-enable the tap and hope we act faster next time
                    CGEventTapEnable(self._tap, True)
                    plover.log.warning(
                        "Keystrokes may have been missed. "
                        + "Keyboard event tap has been re-enabled. ")
                return SUPPRESS_EVENT

            # Don't intercept the event if it has modifiers, allow
            # Fn and Numeric flags so we can suppress the arrow and
            # extended (home, end, etc...) keys.
            suppressible_modifiers = (kCGEventFlagMaskNumericPad |
                                      kCGEventFlagMaskSecondaryFn |
                                      kCGEventFlagMaskNonCoalesced)
            has_nonsupressible_modifiers = \
                CGEventGetFlags(event) & ~suppressible_modifiers
            if has_nonsupressible_modifiers:
                return PASS_EVENT_THROUGH

            keycode = CGEventGetIntegerValueField(
                event, kCGKeyboardEventKeycode)
            key = KEYCODE_TO_KEY.get(keycode)
            self._async_dispatch(key, event_type)
            if key in self._suppressed_keys:
                return SUPPRESS_EVENT
            return PASS_EVENT_THROUGH

        self._tap = CGEventTapCreate(
            kCGSessionEventTap,
            kCGHeadInsertEventTap,
            kCGEventTapOptionDefault,
            CGEventMaskBit(kCGEventKeyDown) | CGEventMaskBit(kCGEventKeyUp),
            callback, None)
        if self._tap is None:
            # Todo(hesky): See if there is a nice way to show the user what's
            # needed (or do it for them).
            raise Exception("Enable access for assistive devices.")
        CGEventTapEnable(self._tap, False)

    def run(self):
        source = CFMachPortCreateRunLoopSource(None, self._tap, 0)
        handler_thread = threading.Thread(
            target=self._event_handler,
            name="KeyEventDispatcher")
        handler_thread.start()
        self._loop = CFRunLoopGetCurrent()
        CFRunLoopAddSource(
            self._loop,
            source,
            kCFRunLoopCommonModes
        )
        CGEventTapEnable(self._tap, True)

        CFRunLoopRun()

        # Wake up event handler.
        self._event_queue.put_nowait(None)
        handler_thread.join()
        CFMachPortInvalidate(self._tap)
        CFRelease(self._tap)
        CFRunLoopSourceInvalidate(source)

    def cancel(self):
        CFRunLoopStop(self._loop)
        self.join()
        self._loop = None

    def suppress_keyboard(self, suppressed_keys=()):
        self._suppressed_keys = set(suppressed_keys)

    def _async_dispatch(self, key, event_type):
        """
        Dispatches a key string in KEYCODE_TO_KEY.values() and a CGEventType
        to the appropriate KeyboardCapture callback
        without blocking execution of its caller.
        """
        if key is None:
            return

        is_keyup = event_type == kCGEventKeyUp
        pair = (key, is_keyup)
        self._event_queue.put_nowait(pair)

    def _event_handler(self):
        """
        Event dispatching thread launched during run().
        Loops until None is received from _event_queue.
        Avoids busy-waiting by blocking on _event_queue.

        In normal operation, it gets a pair of
        (key_string, is_keyup_bool) from _event_queue
        and routes the string to self.key_up or self.key_down,
        then waits for a new pair to arrive.
        """
        while True:
            pair = self._event_queue.get(block=True, timeout=None)
            if pair is None:
                return

            key, is_keyup = pair
            handler = self.key_up if is_keyup else self.key_down
            handler(key)


class KeyboardEmulation:

    RAW_PRESS, STRING_PRESS = range(2)

    def __init__(self):
        self._layout = KeyboardLayout()

    @staticmethod
    def send_backspaces(number_of_backspaces):
        for _ in range(number_of_backspaces):
            backspace_down = CGEventCreateKeyboardEvent(
                OUTPUT_SOURCE, BACK_SPACE, True)
            backspace_up = CGEventCreateKeyboardEvent(
                OUTPUT_SOURCE, BACK_SPACE, False)
            CGEventPost(kCGSessionEventTap, backspace_down)
            CGEventPost(kCGSessionEventTap, backspace_up)

    def send_string(self, s):
        """

        Args:
            s: The string to emulate.

        We can send keys by keycodes or by SetUnicodeString.
        Setting the string is less ideal, but necessary for things like emoji.
        We want to try to group modifier presses, where convenient.
        So, a string like 'THIS dog [dog emoji]' might be processed like:
            'Raw: Shift down t h i s shift up,
            Raw: space d o g space,
            String: [dog emoji]'
        There are 3 groups, the shifted group, the spaces and dog string,
        and the emoji.
        """
        # Key plan will store the type of output
        # (raw keycodes versus setting string)
        # and the list of keycodes or the goal character.
        key_plan = []

        # apply_raw's properties are used to store
        # the current keycode sequence,
        # and add the sequence to the key plan when called as a function.
        def apply_raw():
            if hasattr(apply_raw, 'sequence') \
                    and len(apply_raw.sequence) is not 0:
                apply_raw.sequence.extend(apply_raw.release_modifiers)
                key_plan.append((self.RAW_PRESS, apply_raw.sequence))
            apply_raw.sequence = []
            apply_raw.release_modifiers = []
        apply_raw()

        last_modifier = None
        for c in s:
            for keycode, modifier in self._layout.char_to_key_sequence(c):
                if keycode is not None:
                    if modifier is not last_modifier:
                        # Flush on modifier change.
                        apply_raw()
                        last_modifier = modifier
                        modifier_keycodes = self._modifier_to_keycodes(
                            modifier)
                        apply_raw.sequence.extend(down(modifier_keycodes))
                        apply_raw.release_modifiers.extend(
                            up(modifier_keycodes))
                    apply_raw.sequence.extend(down_up([keycode]))
                else:
                    # Flush on type change.
                    last_modifier = None
                    apply_raw()
                    key_plan.append((self.STRING_PRESS, c))
        # Flush after string is complete.
        apply_raw()

        # We have a key plan for the whole string, grouping modifiers.
        for press_type, sequence in key_plan:
            if press_type is self.STRING_PRESS:
                self._send_string_press(sequence)
            elif press_type is self.RAW_PRESS:
                self._send_sequence(sequence)

    @staticmethod
    def _send_string_press(c):
        event = CGEventCreateKeyboardEvent(OUTPUT_SOURCE, 0, True)
        KeyboardEmulation._set_event_string(event, c)
        CGEventPost(kCGSessionEventTap, event)
        event = CGEventCreateKeyboardEvent(OUTPUT_SOURCE, 0, False)
        KeyboardEmulation._set_event_string(event, c)
        CGEventPost(kCGSessionEventTap, event)

    def send_key_combination(self, combo_string):
        """Emulate a sequence of key combinations.

        Args:
            combo_string: A string representing a sequence of key
                combinations. Keys are represented by their names in the
                Xlib.XK module, without the 'XK_' prefix. For example, the
                left Alt key is represented by 'Alt_L'. Keys are either
                separated by a space or a left or right parenthesis.
                Parentheses must be properly formed in pairs and may be
                nested. A key immediately followed by a parenthetical
                indicates that the key is pressed down while all keys enclosed
                in the parenthetical are pressed and released in turn. For
                example, Alt_L(Tab) means to hold the left Alt key down, press
                and release the Tab key, and then release the left Alt key.

        """
        def name_to_code(name):
            # Static key codes
            code = KEYNAME_TO_KEYCODE.get(name)
            if code is not None:
                pass
            # Dead keys
            elif name.startswith('dead_'):
                code, mod = self._layout.deadkey_symbol_to_key_sequence(
                    DEADKEY_SYMBOLS.get(name)
                )[0]
            # Normal keys
            else:
                char = KEYNAME_TO_CHAR.get(name, name)
                code, mods = self._layout.char_to_key_sequence(char)[0]
            return code
        # Parse and validate combo.
        key_events = parse_key_combo(combo_string, name_to_code)
        # Send events...
        self._send_sequence(key_events)

    @staticmethod
    def _modifier_to_keycodes(modifier):
        keycodes = []
        if modifier & 16:
            keycodes.append(KEYNAME_TO_KEYCODE['control_r'])
        if modifier & 8:
            keycodes.append(KEYNAME_TO_KEYCODE['alt_r'])
        if modifier & 2:
            keycodes.append(KEYNAME_TO_KEYCODE['shift_r'])
        if modifier & 1:
            keycodes.append(KEYNAME_TO_KEYCODE['super_r'])
        return keycodes

    @staticmethod
    def _set_event_string(event, s):
        nb_utf16_codepoints = len(s.encode('utf-16-le')) // 2
        CGEventKeyboardSetUnicodeString(event, nb_utf16_codepoints, s)

    MODS_MASK = (
        kCGEventFlagMaskAlternate |
        kCGEventFlagMaskControl |
        kCGEventFlagMaskShift |
        kCGEventFlagMaskCommand |
        kCGEventFlagMaskSecondaryFn
    )

    @staticmethod
    def _get_media_event(key_id, key_down):
        # Credit: https://gist.github.com/fredrikw/4078034
        flags = 0xa00 if key_down else 0xb00
        return NSEvent.otherEventWithType_location_modifierFlags_timestamp_windowNumber_context_subtype_data1_data2_(  # nopep8: We can't make this line shorter!
            NSSystemDefined, (0, 0),
            flags,
            0, 0, 0, 8,
            (key_id << 16) | flags,
            -1
        ).CGEvent()

    @staticmethod
    def _send_sequence(sequence):
        # There is a bug in the event system that seems to cause inconsistent
        # modifiers on key events:
        # http://stackoverflow.com/questions/2008126/cgeventpost-possible-bug-when-simulating-keyboard-events
        # My solution is to manage the state myself.
        # I'm not sure how to deal with caps lock.
        # If mods_flags is not zero at the end then bad things might happen.
        mods_flags = 0

        for keycode, key_down in sequence:
            if keycode >= NX_KEY_OFFSET:
                # Handle media (NX) key.
                event = KeyboardEmulation._get_media_event(
                    keycode - NX_KEY_OFFSET, key_down)
            else:
                # Handle regular keycode.
                if not key_down and keycode in MODIFIER_KEYS_TO_MASKS:
                    mods_flags &= ~MODIFIER_KEYS_TO_MASKS[keycode]

                if key_down and keycode_needs_fn_mask(keycode):
                    mods_flags |= kCGEventFlagMaskSecondaryFn

                event = CGEventCreateKeyboardEvent(
                    OUTPUT_SOURCE, keycode, key_down)

                if key_down and keycode not in MODIFIER_KEYS_TO_MASKS:
                    event_flags = CGEventGetFlags(event)
                    # Add wanted flags, remove unwanted flags.
                    goal_flags = ((event_flags & ~KeyboardEmulation.MODS_MASK)
                                  | mods_flags)
                    if event_flags != goal_flags:
                        CGEventSetFlags(event, goal_flags)

                    # Half millisecond pause after key down.
                    sleep(0.0005)

                if key_down and keycode in MODIFIER_KEYS_TO_MASKS:
                    mods_flags |= MODIFIER_KEYS_TO_MASKS[keycode]

                if not key_down and keycode_needs_fn_mask(keycode):
                    mods_flags &= ~kCGEventFlagMaskSecondaryFn

            CGEventPost(kCGSessionEventTap, event)



# Copyright (c) 2013 Hesky Fisher.
# See LICENSE.txt for details.

"""StenoDictionary class and related functions.

A steno dictionary maps sequences of steno strokes to translations.

"""

import collections
import os
import shutil

from plover.resource import ASSET_SCHEME, resource_filename, resource_timestamp


class StenoDictionary:
    """A steno dictionary.

    This dictionary maps immutable sequences to translations and tracks the
    length of the longest key.

    Attributes:
    longest_key -- A read only property holding the length of the longest key.
    timestamp -- File last modification time, used to detect external changes.

    """

    # False if class support creation.
    readonly = False

    def __init__(self):
        self._dict = {}
        self._longest_key_length = 0
        self._longest_listener_callbacks = set()
        self.reverse = collections.defaultdict(list)
        # Case-insensitive reverse dict
        self.casereverse = collections.defaultdict(collections.Counter)
        self.filters = []
        self.timestamp = 0
        self.readonly = False
        self.enabled = True
        self.path = None

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self.path)

    def __repr__(self):
        return str(self)

    @classmethod
    def create(cls, resource):
        assert not resource.startswith(ASSET_SCHEME)
        if cls.readonly:
            raise ValueError('%s does not support creation' % cls.__name__)
        d = cls()
        d.path = resource
        return d

    @classmethod
    def load(cls, resource):
        filename = resource_filename(resource)
        timestamp = resource_timestamp(filename)
        d = cls()
        d._load(filename)
        if resource.startswith(ASSET_SCHEME) or \
           not os.access(filename, os.W_OK):
            d.readonly = True
        d.path = resource
        d.timestamp = timestamp
        return d

    def save(self):
        assert not self.readonly
        filename = resource_filename(self.path)
        # Write the new file to a temp location.
        tmp = filename + '.tmp'
        self._save(tmp)
        timestamp = resource_timestamp(tmp)
        # Then move the new file to the final location.
        shutil.move(tmp, filename)
        # And update our timestamp.
        self.timestamp = timestamp

    def _load(self, filename):
        raise NotImplementedError()

    def _save(self, filename):
        raise NotImplementedError()

    @property
    def longest_key(self):
        """The length of the longest key in the dict."""
        return self._longest_key

    def __len__(self):
        return self._dict.__len__()

    def __iter__(self):
        return self._dict.__iter__()

    def __getitem__(self, key):
        return self._dict.__getitem__(key)

    def clear(self):
        self._dict.clear()
        self.reverse.clear()
        self.casereverse.clear()
        self._longest_key = 0

    def items(self):
        return self._dict.items()

    def update(self, *args, **kwargs):
        assert not self.readonly
        longest_key = 0
        _dict = self._dict
        reverse = self.reverse
        casereverse = self.casereverse
        for iterable in args + (kwargs,):
            if isinstance(iterable, (dict, StenoDictionary)):
                iterable = iterable.items()
            for key, value in iterable:
                longest_key = max(longest_key, len(key))
                _dict[key] = value
                reverse[value].append(key)
                casereverse[value.lower()][value] += 1
        self._longest_key = max(self._longest_key, longest_key)

    def __setitem__(self, key, value):
        assert not self.readonly
        self._longest_key = max(self._longest_key, len(key))
        self._dict[key] = value
        self.reverse[value].append(key)
        self.casereverse[value.lower()][value] += 1

    def get(self, key, fallback=None):
        return self._dict.get(key, fallback)

    def __delitem__(self, key):
        assert not self.readonly
        value = self._dict.pop(key)
        self.reverse[value].remove(key)
        counter = self.casereverse[value.lower()]
        count = counter.pop(value) - 1
        if count:
            counter[value] = count
        if len(key) == self.longest_key:
            if self._dict:
                self._longest_key = max(len(x) for x in self._dict)
            else:
                self._longest_key = 0

    def __contains__(self, key):
        return self.get(key) is not None

    def reverse_lookup(self, value):
        return self.reverse[value]

    def casereverse_lookup(self, value):
        return list(self.casereverse[value].keys())

    @property
    def _longest_key(self):
        return self._longest_key_length

    @_longest_key.setter
    def _longest_key(self, longest_key):
        if longest_key == self._longest_key_length:
            return
        self._longest_key_length = longest_key
        for callback in self._longest_listener_callbacks:
            callback(longest_key)

    def add_longest_key_listener(self, callback):
        self._longest_listener_callbacks.add(callback)

    def remove_longest_key_listener(self, callback):
        self._longest_listener_callbacks.remove(callback)


class StenoDictionaryCollection:

    def __init__(self, dicts=[]):
        self.dicts = []
        self.filters = []
        self.longest_key = 0
        self.longest_key_callbacks = set()
        self.set_dicts(dicts)

    def set_dicts(self, dicts):
        for d in self.dicts:
            d.remove_longest_key_listener(self._longest_key_listener)
        self.dicts = dicts[:]
        for d in self.dicts:
            d.add_longest_key_listener(self._longest_key_listener)
        self._longest_key_listener()

    def _lookup(self, key, dicts=None, filters=()):
        if dicts is None:
            dicts = self.dicts
        key_len = len(key)
        if key_len > self.longest_key:
            return None
        for d in dicts:
            if not d.enabled:
                continue
            if key_len > d.longest_key:
                continue
            value = d.get(key)
            if value:
                for f in filters:
                    if f(key, value):
                        return None
                return value

    def __str__(self):
        return 'StenoDictionaryCollection' + repr(tuple(self.dicts))

    def __repr__(self):
        return str(self)

    def lookup(self, key):
        return self._lookup(key, filters=self.filters)

    def raw_lookup(self, key):
        return self._lookup(key)

    def reverse_lookup(self, value):
        keys = []
        for n, d in enumerate(self.dicts):
            if not d.enabled:
                continue
            for k in d.reverse_lookup(value):
                # Ignore key if it's overridden by a higher priority dictionary.
                if self._lookup(k, dicts=self.dicts[:n]) is None:
                    keys.append(k)
        return keys

    def casereverse_lookup(self, value):
        for d in self.dicts:
            if not d.enabled:
                continue
            key = d.casereverse_lookup(value)
            if key:
                return key

    def first_writable(self):
        '''Return the first writable dictionary.'''
        for d in self.dicts:
            if not d.readonly:
                return d
        raise KeyError('no writable dictionary')

    def set(self, key, value, path=None):
        if path is None:
            d = self.first_writable()
        else:
            d = self[path]
        d[key] = value

    def save(self, path_list=None):
        '''Save the dictionaries in <path_list>.

        If <path_list> is None, all writable dictionaries are saved'''
        if path_list is None:
            dict_list = [d for d in self if dictionary.save is not None]
        else:
            dict_list = [self[path] for path in path_list]
        for d in dict_list:
            d.save()

    def get(self, path):
        for d in self.dicts:
            if d.path == path:
                return d

    def __getitem__(self, path):
        d = self.get(path)
        if d is None:
            raise KeyError(repr(path))
        return d

    def __iter__(self):
        for d in self.dicts:
            yield d.path

    def add_filter(self, f):
        self.filters.append(f)

    def remove_filter(self, f):
        self.filters.remove(f)

    def add_longest_key_listener(self, callback):
        self.longest_key_callbacks.add(callback)

    def remove_longest_key_listener(self, callback):
        self.longest_key_callbacks.remove(callback)
    
    def _longest_key_listener(self, ignored=None):
        if self.dicts:
            new_longest_key = max(d.longest_key for d in self.dicts)
        else:
            new_longest_key = 0
        if new_longest_key != self.longest_key:
            self.longest_key = new_longest_key
            for c in self.longest_key_callbacks:
                c(new_longest_key)




from collections import namedtuple

import pkg_resources

from plover.oslayer.config import HAS_GUI_QT, PLUGINS_PLATFORM
from plover import log


class Plugin:

    def __init__(self, plugin_type, name, obj):
        self.plugin_type = plugin_type
        self.name = name
        self.obj = obj
        self.__doc__ = obj.__doc__ or ''

    def __str__(self):
        return '%s:%s' % (self.plugin_type, self.name)


PluginDistribution = namedtuple('PluginDistribution', 'dist plugins')


class Registry:

    PLUGIN_TYPES = (
        'command',
        'dictionary',
        'extension',
        'gui',
        'gui.qt.machine_option',
        'gui.qt.tool',
        'machine',
        'macro',
        'meta',
        'system',
    )

    def __init__(self, suppress_errors=True):
        self._plugins = {}
        self._distributions = {}
        self._suppress_errors = suppress_errors
        for plugin_type in self.PLUGIN_TYPES:
            self._plugins[plugin_type] = {}

    def register_plugin(self, plugin_type, name, obj):
        plugin = Plugin(plugin_type, name, obj)
        self._plugins[plugin_type][name.lower()] = plugin
        return plugin

    def register_plugin_from_entrypoint(self, plugin_type, entrypoint):
        log.info('%s: %s (from %s in %s)', plugin_type, entrypoint.name,
                 entrypoint.dist, entrypoint.dist.location)
        try:
            obj = entrypoint.load()
        except:
            log.error('error loading %s plugin: %s (from %s)', plugin_type,
                      entrypoint.name, entrypoint.module_name, exc_info=True)
            if not self._suppress_errors:
                raise
        else:
            plugin = self.register_plugin(plugin_type, entrypoint.name, obj)
            # Keep track of distributions providing plugins.
            dist_id = str(entrypoint.dist)
            dist = self._distributions.get(dist_id)
            if dist is None:
                dist = PluginDistribution(entrypoint.dist, set())
                self._distributions[dist_id] = dist
            dist.plugins.add(plugin)

    def get_plugin(self, plugin_type, plugin_name):
        return self._plugins[plugin_type][plugin_name.lower()]

    def list_plugins(self, plugin_type):
        return sorted(self._plugins[plugin_type].values(),
                      key=lambda p: p.name)

    def list_distributions(self):
        return [dist for dist_id, dist in sorted(self._distributions.items())]

    def update(self):
        for plugin_type in self.PLUGIN_TYPES:
            if plugin_type.startswith('gui.qt.') and not HAS_GUI_QT:
                continue
            entrypoint_type = 'plover.%s' % plugin_type
            for entrypoint in pkg_resources.iter_entry_points(entrypoint_type):
                if 'gui_qt' in entrypoint.extras and not HAS_GUI_QT:
                    continue
                self.register_plugin_from_entrypoint(plugin_type, entrypoint)
            if PLUGINS_PLATFORM is not None:
                entrypoint_type = 'plover.%s.%s' % (PLUGINS_PLATFORM, plugin_type)
                for entrypoint in pkg_resources.iter_entry_points(entrypoint_type):
                    self.register_plugin_from_entrypoint(plugin_type, entrypoint)


registry = Registry()





import os
import logging

import dbus

from plover import log, __name__ as __software_name__
from plover.oslayer.config import ASSETS_DIR


APPNAME = __software_name__.capitalize()
APPICON = os.path.join(ASSETS_DIR, 'plover.png')
SERVICE = 'org.freedesktop.Notifications'
INTERFACE = '/org/freedesktop/Notifications'


class DbusNotificationHandler(logging.Handler):
    """ Handler using DBus notifications to show messages. """

    def __init__(self):
        super().__init__()
        self._bus = dbus.SessionBus()
        self._proxy = self._bus.get_object(SERVICE, INTERFACE)
        self._notify = dbus.Interface(self._proxy, SERVICE)
        self.setLevel(log.WARNING)
        self.setFormatter(log.NoExceptionTracebackFormatter('<b>%(levelname)s:</b> %(message)s'))

    def emit(self, record):
        level = record.levelno
        message = self.format(record)
        if message.endswith('\n'):
            message = message[:-1]
        if level <= log.INFO:
            timeout = 10
            urgency = 0
        elif level <= log.WARNING:
            timeout = 15
            urgency = 1
        else:
            timeout = 0
            urgency = 2
        self._notify.Notify(APPNAME, 0, APPICON,  # app_name, replaces_id, app_icon
                            APPNAME, message, '', # actions
                            { 'urgency': dbus.Byte(urgency) },
                            timeout * 1000)




import os
import sys
import logging

from plyer import notification

from plover import log, __name__ as __software_name__
from plover.oslayer.config import ASSETS_DIR


APPNAME = __software_name__.capitalize()

if sys.platform.startswith('win32'):
    APPICON = os.path.join(ASSETS_DIR, 'plover.ico')
else:
    APPICON = os.path.join(ASSETS_DIR, 'plover_32x32.png')


class PlyerNotificationHandler(logging.Handler):
    """ Handler using Plyer's notifications to show messages. """

    def __init__(self):
        super().__init__()
        self.setLevel(log.WARNING)
        self.setFormatter(log.NoExceptionTracebackFormatter('%(levelname)s: %(message)s'))

    def emit(self, record):
        level = record.levelno
        message = self.format(record)
        if message.endswith('\n'):
            message = message[:-1]
        if level <= log.INFO:
            timeout = 10
        elif level <= log.WARNING:
            timeout = 15
        else:
            timeout = 60
        notification.notify(
            app_name=APPNAME,
            app_icon=APPICON,
            title=APPNAME,
            message=message,
            timeout=timeout,
        )




import re

from PyQt5.QtCore import Qt
from PyQt5.QtGui import (
    QCursor,
    QFont,
)
from PyQt5.QtWidgets import (
    QAction,
    QFontDialog,
    QMenu,
)

from plover.suggestions import Suggestion
from plover.formatting import RetroFormatter

from plover.gui_qt.suggestions_dialog_ui import Ui_SuggestionsDialog
from plover.gui_qt.i18n import get_gettext
from plover.gui_qt.utils import ToolBar
from plover.gui_qt.tool import Tool


_ = get_gettext()


class SuggestionsDialog(Tool, Ui_SuggestionsDialog):

    ''' Suggest possible strokes for the last written words. '''

    TITLE = _('Suggestions')
    ICON = ':/lightbulb.svg'
    ROLE = 'suggestions'
    SHORTCUT = 'Ctrl+J'

    WORD_RX = re.compile(r'(?:\w+|[^\w\s]+)\s*')

    STYLE_TRANSLATION, STYLE_STROKES = range(2)

    # Anatomy of the text document:
    # - "root" frame:
    #  - 0+ "suggestions" frames
    #   - 1+ "translation" frames
    #    - 1-10 "strokes" frames

    def __init__(self, engine):
        super().__init__(engine)
        self.setupUi(self)
        self._last_suggestions = None
        # Toolbar.
        self.layout().addWidget(ToolBar(
            self.action_ToggleOnTop,
            self.action_SelectFont,
            self.action_Clear,
        ))
        self.action_Clear.setEnabled(False)
        # Font popup menu.
        self._font_menu = QMenu()
        self._font_menu_text = QAction(_('&Text'), self._font_menu)
        self._font_menu_strokes = QAction(_('&Strokes'), self._font_menu)
        self._font_menu.addActions([self._font_menu_text, self._font_menu_strokes])
        engine.signal_connect('translated', self.on_translation)
        self.suggestions.setFocus()
        self.restore_state()
        self.finished.connect(self.save_state)

    def _get_font(self, name):
        return getattr(self.suggestions, name)

    def _set_font(self, name, font):
        setattr(self.suggestions, name, font)

    def _restore_state(self, settings):
        for name in (
            'text_font',
            'strokes_font',
        ):
            font_string = settings.value(name)
            if font_string is None:
                continue
            font = QFont()
            if not font.fromString(font_string):
                continue
            self._set_font(name, font)
        ontop = settings.value('ontop', None, bool)
        if ontop is not None:
            self.action_ToggleOnTop.setChecked(ontop)
            self.on_toggle_ontop(ontop)

    def _save_state(self, settings):
        for name in (
            'text_font',
            'strokes_font',
        ):
            font = self._get_font(name)
            font_string = font.toString()
            settings.setValue(name, font_string)
        ontop = bool(self.windowFlags() & Qt.WindowStaysOnTopHint)
        settings.setValue('ontop', ontop)

    def _show_suggestions(self, suggestion_list):
        self.suggestions.append(suggestion_list)
        self.action_Clear.setEnabled(True)

    @staticmethod
    def tails(ls):
        ''' Return all tail combinations (a la Haskell)

            tails :: [x] -> [[x]]
            >>> tails('abcd')
            ['abcd', 'bcd', 'cd', d']

        '''

        for i in range(len(ls)):
            yield ls[i:]

    def on_translation(self, old, new):

        # Check for new output.
        for a in reversed(new):
            if a.text and not a.text.isspace():
                break
        else:
            return

        # Get the last 10 words.
        with self._engine:
            last_translations = self._engine.translator_state.translations
            retro_formatter = RetroFormatter(last_translations)
            split_words = retro_formatter.last_words(10, rx=self.WORD_RX)

        suggestion_list = []
        for phrase in self.tails(split_words):
            phrase = ''.join(phrase)
            suggestion_list.extend(self._engine.get_suggestions(phrase))

        if not suggestion_list and split_words:
            suggestion_list = [Suggestion(split_words[-1], [])]

        if suggestion_list and suggestion_list != self._last_suggestions:
            self._last_suggestions = suggestion_list
            self._show_suggestions(suggestion_list)

    def on_select_font(self):
        action = self._font_menu.exec_(QCursor.pos())
        if action is None:
            return
        if action == self._font_menu_text:
            name = 'text_font'
            font_options = ()
        elif action == self._font_menu_strokes:
            name = 'strokes_font'
            font_options = (QFontDialog.MonospacedFonts,)
        font = self._get_font(name)
        font, ok = QFontDialog.getFont(font, self, '', *font_options)
        if ok:
            self._set_font(name, font)

    def on_toggle_ontop(self, ontop):
        flags = self.windowFlags()
        if ontop:
            flags |= Qt.WindowStaysOnTopHint
        else:
            flags &= ~Qt.WindowStaysOnTopHint
        self.setWindowFlags(flags)
        self.show()

    def on_clear(self):
        self.action_Clear.setEnabled(False)
        self._last_suggestions = None
        self.suggestions.clear()



from plover.gui_qt.add_translation_dialog_ui import Ui_AddTranslationDialog
from plover.gui_qt.i18n import get_gettext
from plover.gui_qt.tool import Tool


_ = get_gettext()


class AddTranslationDialog(Tool, Ui_AddTranslationDialog):

    ''' Add a new translation to the dictionary. '''

    TITLE = _('Add Translation')
    ICON = ':/translation_add.svg'
    ROLE = 'add_translation'
    SHORTCUT = 'Ctrl+N'

    def __init__(self, engine, dictionary_path=None):
        super().__init__(engine)
        self.setupUi(self)
        self.add_translation.select_dictionary(dictionary_path)
        engine.signal_connect('config_changed', self.on_config_changed)
        self.on_config_changed(engine.config)
        self.installEventFilter(self)
        self.restore_state()
        self.finished.connect(self.save_state)

    def on_config_changed(self, config_update):
        if 'translation_frame_opacity' in config_update:
            opacity = config_update.get('translation_frame_opacity')
            if opacity is None:
                return
            assert 0 <= opacity <= 100
            self.setWindowOpacity(opacity / 100.0)

    def accept(self):
        self.add_translation.save_entry()
        super().accept()

    def reject(self):
        self.add_translation.reject()
        super().reject()



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

# TODO: unit test this file

"""Generic stenography data models.

This module contains the following class:

Stroke -- A data model class that encapsulates a sequence of steno keys.

"""

import re

from plover import system


STROKE_DELIMITER = '/'

_NUMBERS = set('0123456789')
_IMPLICIT_NUMBER_RX = re.compile('(^|[1-4])([6-9])')

def normalize_stroke(stroke):
    letters = set(stroke)
    if letters & _NUMBERS:
        if system.NUMBER_KEY in letters:
            stroke = stroke.replace(system.NUMBER_KEY, '')
        # Insert dash when dealing with 'explicit' numbers
        m = _IMPLICIT_NUMBER_RX.search(stroke)
        if m is not None:
            start = m.start(2)
            return stroke[:start] + '-' + stroke[start:]
    if '-' in letters:
        if stroke.endswith('-'):
            stroke = stroke[:-1]
        elif letters & system.IMPLICIT_HYPHENS:
            stroke = stroke.replace('-', '')
    return stroke

def normalize_steno(strokes_string):
    """Convert steno strings to one common form."""
    return tuple(normalize_stroke(stroke) for stroke
                 in strokes_string.split(STROKE_DELIMITER))

def sort_steno_keys(steno_keys):
    return sorted(steno_keys, key=lambda x: system.KEY_ORDER.get(x, -1))

def sort_steno_strokes(strokes_list):
    '''Return suggestions, sorted by fewest strokes, then fewest keys.'''
    return sorted(strokes_list, key=lambda x: (len(x), sum(map(len, x))))


class Stroke:
    """A standardized data model for stenotype machine strokes.

    This class standardizes the representation of a stenotype chord. A stenotype
    chord can be any sequence of stenotype keys that can be simultaneously
    pressed. Nearly all stenotype machines offer the same set of keys that can
    be combined into a chord, though some variation exists due to duplicate
    keys. This class accounts for such duplication, imposes the standard
    stenographic ordering on the keys, and combines the keys into a single
    string (called RTFCRE for historical reasons).

    """

    def __init__(self, steno_keys) :
        """Create a steno stroke by formatting steno keys.

        Arguments:

        steno_keys -- A sequence of pressed keys.

        """
        # Remove duplicate keys and save local versions of the input 
        # parameters.
        steno_keys_set = set(steno_keys)
        # Order the steno keys so comparisons can be made.
        steno_keys = list(sort_steno_keys(steno_keys_set))

        # Convert strokes involving the number bar to numbers.
        if system.NUMBER_KEY in steno_keys:
            numeral = False
            for i, e in enumerate(steno_keys):
                if e in system.NUMBERS:
                    steno_keys[i] = system.NUMBERS[e]
                    numeral = True
            if numeral:
                steno_keys.remove(system.NUMBER_KEY)
        
        if steno_keys_set & system.IMPLICIT_HYPHEN_KEYS:
            self.rtfcre = ''.join(key.strip('-') for key in steno_keys)
        else:
            pre = ''.join(k.strip('-') for k in steno_keys if k[-1] == '-' or 
                          k == system.NUMBER_KEY)
            post = ''.join(k.strip('-') for k in steno_keys if k[0] == '-')
            self.rtfcre = '-'.join([pre, post]) if post else pre

        self.steno_keys = steno_keys

        # Determine if this stroke is a correction stroke.
        self.is_correction = (self.rtfcre == system.UNDO_STROKE_STENO)

    def __str__(self):
        if self.is_correction:
            prefix = '*'
        else:
            prefix = ''
        return '%sStroke(%s : %s)' % (prefix, self.rtfcre, self.steno_keys)

    def __eq__(self, other):
        return (isinstance(other, Stroke)
                and self.steno_keys == other.steno_keys)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return str(self)




# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Parsing a json formatted dictionary.

"""

import codecs

try:
    import simplejson as json
except ImportError:
    import json

from plover.steno_dictionary import StenoDictionary
from plover.steno import normalize_steno


class JsonDictionary(StenoDictionary):

    def _load(self, filename):
        with open(filename, 'rb') as fp:
            contents = fp.read()
        for encoding in ('utf-8', 'latin-1'):
            try:
                contents = contents.decode(encoding)
            except UnicodeDecodeError:
                continue
            else:
                break
        else:
            raise ValueError('\'%s\' encoding could not be determined' % (filename,))
        d = dict(json.loads(contents))
        self.update((normalize_steno(x[0]), x[1]) for x in d.items())

    def _save(self, filename):
        with open(filename, 'wb') as fp:
            writer = codecs.getwriter('utf-8')(fp)
            json.dump({'/'.join(k): v for k, v in self.items()},
                      writer, ensure_ascii=False, sort_keys=True,
                      indent=0, separators=(',', ': '))



from threading import Event

from plover.oslayer.keyboardcontrol import KeyboardEmulation

from plover.gui_none.engine import Engine


def show_error(title, message):
    print('%s: %s' % (title, message))


def main(config):
    engine = Engine(config, KeyboardEmulation())
    if not engine.load_config():
        return 3
    quitting = Event()
    engine.hook_connect('quit', quitting.set)
    engine.start()
    try:
        quitting.wait()
    except KeyboardInterrupt:
        engine.quit()
    return engine.join()



# -*- coding: utf-8 -*-
# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"Launch the plover application."


import argparse
import atexit
import os
import sys
import subprocess
import traceback

import pkg_resources

if sys.platform.startswith('darwin'):
    import appnope

from plover.config import CONFIG_DIR, CONFIG_FILE, Config
from plover.oslayer import processlock
from plover.registry import registry
from plover import log
from plover import __name__ as __software_name__
from plover import __version__


def init_config_dir():
    """Creates plover's config dir.

    This usually only does anything the first time plover is launched.
    """
    # Create the configuration directory if needed.
    if not os.path.exists(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)

    # Create a default configuration file if one doesn't already exist.
    if not os.path.exists(CONFIG_FILE):
        with open(CONFIG_FILE, 'wb') as f:
            f.close()


def main():
    """Launch plover."""
    description = "Run the plover stenotype engine. This is a graphical application."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--version', action='version', version='%s %s'
                        % (__software_name__.capitalize(), __version__))
    parser.add_argument('-s', '--script', default=None, nargs=argparse.REMAINDER,
                        help='use another plugin console script as main entrypoint, '
                        'passing in the rest of the command line arguments, '
                        'print list of available scripts when no argument is given')
    parser.add_argument('-l', '--log-level', choices=['debug', 'info', 'warning', 'error'],
                        default=None, help='set log level')
    parser.add_argument('-g', '--gui', default=None, help='set gui')
    args = parser.parse_args(args=sys.argv[1:])
    if args.log_level is not None:
        log.set_level(args.log_level.upper())
    log.setup_platform_handler()

    log.info('Plover %s', __version__)
    log.info('configuration directory: %s', CONFIG_DIR)

    registry.update()

    if args.gui is None:
        gui_priority = {
            'qt': 1,
            'none': -1,
        }
        gui_list = sorted(registry.list_plugins('gui'), reverse=True,
                          key=lambda gui: gui_priority.get(gui.name, 0))
        gui = gui_list[0].obj
    else:
        gui = registry.get_plugin('gui', args.gui).obj

    try:
        if args.script is not None:
            if args.script:
                # Create a mapping of available console script,
                # with the following priorities (highest first):
                # - {project_name}-{version}:{script_name}
                # - {project_name}:{script_name}
                # - {script_name}
                console_scripts = {}
                for e in sorted(pkg_resources.iter_entry_points('console_scripts'),
                                key=lambda e: (e.dist, e.name)):
                    for key in (
                        '%s-%s:%s' % (e.dist.project_name, e.dist.version, e.name),
                        '%s:%s' % (e.dist.project_name, e.name),
                        e.name,
                    ):
                        console_scripts[key] = e
                entrypoint = console_scripts.get(args.script[0])
                if entrypoint is None:
                    log.error('no such script: %s', args.script[0])
                    code = 1
                else:
                    sys.argv = args.script
                    try:
                        code = entrypoint.load()()
                    except SystemExit as e:
                        code = e.code
                if code is None:
                    code = 0
            else:
                print('available script(s):')
                dist = None
                for e in sorted(pkg_resources.iter_entry_points('console_scripts'),
                                key=lambda e: (str(e.dist), e.name)):
                    if dist != e.dist:
                        dist = e.dist
                        print('%s:' % dist)
                    print('- %s' % e.name)
                code = 0
            os._exit(code)

        # Ensure only one instance of Plover is running at a time.
        with processlock.PloverLock():
            if sys.platform.startswith('darwin'):
                appnope.nope()
            init_config_dir()
            # This must be done after calling init_config_dir, so
            # Plover's configuration directory actually exists.
            log.setup_logfile()
            config = Config()
            config.target_file = CONFIG_FILE
            code = gui.main(config)
            with open(config.target_file, 'wb') as f:
                config.save(f)
    except processlock.LockNotAcquiredException:
        gui.show_error('Error', 'Another instance of Plover is already running.')
        code = 1
    except:
        gui.show_error('Unexpected error', traceback.format_exc())
        code = 2
    if code == -1:
        # Restart.
        args = sys.argv[:]
        if args[0].endswith('.py') or args[0].endswith('.pyc'):
            # We're running from source.
            assert args[0] == __file__
            args[0:1] = [sys.executable, '-m', __spec__.name]
        # Execute atexit handlers.
        atexit._run_exitfuncs()
        if sys.platform.startswith('win32'):
            # Workaround https://bugs.python.org/issue19066
            subprocess.Popen(args, cwd=os.getcwd())
            code = 0
        else:
            os.execv(args[0], args)
    os._exit(code)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import glob
import shutil
import sys
import os


def trim(directory, patterns_file, verbose=True):
    # Build list of patterns.
    pattern_list = []
    subdir = directory
    with open(patterns_file) as fp:
        for line in fp:
            line = line.strip()
            # Empty line, ignore.
            if not line:
                continue
            # Comment, ignore.
            if line.startswith('#'):
                continue
            # Sub-directory change.
            if line.startswith(':'):
                subdir = os.path.join(directory, line[1:])
                continue
            # Pattern (relative to current sub-directory).
            pattern_list.append(os.path.join(subdir, line))
    # Trim directory tree.
    for pattern in pattern_list:
        for path in reversed(glob.glob(pattern, recursive=True)):
            if verbose:
                print('removing', path)
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)


if __name__ == '__main__':
    directory, patterns_file = sys.argv[1:]
    trim(directory, patterns_file)




import sys

from plover.translation import Translation


if sys.platform.startswith('darwin'):
    BACK_STRING = '{#Alt_L(BackSpace)}{^}'
else:
    BACK_STRING = '{#Control_L(BackSpace)}{^}'

def undo(translator, stroke, cmdline):
    for t in reversed(translator.get_state().translations):
        translator.untranslate_translation(t)
        if t.has_undo():
            return
    # There is no more buffer to delete from -- remove undo and add a
    # stroke that removes last word on the user's OS, but don't add it
    # to the state history.
    translator.flush([Translation([stroke], BACK_STRING)])




import sys

from PyQt5.QtCore import (
    QThread,
    QVariant,
    pyqtSignal,
)

from plover.engine import StenoEngine


class Engine(StenoEngine, QThread):

    # Signals.
    signal_stroked = pyqtSignal(QVariant)
    signal_translated = pyqtSignal(QVariant, QVariant)
    signal_machine_state_changed = pyqtSignal(str, str)
    signal_output_changed = pyqtSignal(bool)
    signal_config_changed = pyqtSignal(QVariant)
    signal_dictionaries_loaded = pyqtSignal(QVariant)
    signal_send_string = pyqtSignal(str)
    signal_send_backspaces = pyqtSignal(int)
    signal_send_key_combination = pyqtSignal(str)
    signal_add_translation = pyqtSignal()
    signal_focus = pyqtSignal()
    signal_configure = pyqtSignal()
    signal_lookup = pyqtSignal()
    signal_quit = pyqtSignal()

    def __init__(self, config, keyboard_emulation):
        StenoEngine.__init__(self, config, keyboard_emulation)
        QThread.__init__(self)
        self._signals = {}
        for hook in self.HOOKS:
            signal = getattr(self, 'signal_' + hook)
            self.hook_connect(hook, signal.emit)
            self._signals[hook] = signal

    def _in_engine_thread(self):
        return self.currentThread() == self

    def start(self):
        QThread.start(self)
        StenoEngine.start(self)

    def join(self):
        QThread.wait(self)
        return self.code

    def run(self):
        if sys.platform.startswith('darwin'):
            import appnope
            appnope.nope()
        super().run()

    def signal_connect(self, name, callback):
        self._signals[name].connect(callback)



from distutils import log
import contextlib
import importlib
import os
import shutil
import subprocess
import sys

from setuptools.command.build_py import build_py
import pkg_resources
import setuptools


class Command(setuptools.Command):

    def build_in_place(self):
        self.run_command('build_py')
        self.reinitialize_command('build_ext', inplace=1)
        self.run_command('build_ext')

    @contextlib.contextmanager
    def project_on_sys_path(self):
        self.build_in_place()
        ei_cmd = self.get_finalized_command("egg_info")
        old_path = sys.path[:]
        old_modules = sys.modules.copy()
        try:
            sys.path.insert(0, pkg_resources.normalize_path(ei_cmd.egg_base))
            pkg_resources.working_set.__init__()
            pkg_resources.add_activation_listener(lambda dist: dist.activate())
            pkg_resources.require('%s==%s' % (ei_cmd.egg_name, ei_cmd.egg_version))
            yield
        finally:
            sys.path[:] = old_path
            sys.modules.clear()
            sys.modules.update(old_modules)
            pkg_resources.working_set.__init__()

# `test` command. {{{

class Test(Command):

    description = 'run unit tests after in-place build'
    command_consumes_arguments = True
    user_options = []
    test_dir = 'test'

    def initialize_options(self):
        self.args = []

    def finalize_options(self):
        pass

    def run(self):
        with self.project_on_sys_path():
            self.run_tests()

    def run_tests(self):
        # Remove __pycache__ directory so pytest does not freak out
        # when switching between the Linux/Windows versions.
        pycache = os.path.join(self.test_dir, '__pycache__')
        if os.path.exists(pycache):
            shutil.rmtree(pycache)
        custom_testsuite = None
        args = []
        for a in self.args:
            if '-' == a[0]:
                args.append(a)
            elif os.path.exists(a):
                custom_testsuite = a
                args.append(a)
            else:
                args.extend(('-k', a))
        if custom_testsuite is None:
            args.insert(0, self.test_dir)
        sys.argv[1:] = args
        main = pkg_resources.load_entry_point('pytest',
                                              'console_scripts',
                                              'py.test')
        sys.exit(main())

# }}}

# UI generation. {{{

class BuildUi(Command):

    description = 'build UI files'
    user_options = [
        ('force', 'f',
         'force re-generation of all UI files'),
    ]

    hooks = '''
    plover_build_utils.pyqt:fix_icons
    plover_build_utils.pyqt:gettext
    '''.split()

    def initialize_options(self):
        self.force = False

    def finalize_options(self):
        pass

    def _build_ui(self, src):
        dst = os.path.splitext(src)[0] + '_ui.py'
        if not self.force and os.path.exists(dst) and \
           os.path.getmtime(dst) >= os.path.getmtime(src):
            return
        cmd = (
            sys.executable, '-m', 'PyQt5.uic.pyuic',
            '--from-import', src,
        )
        if self.verbose:
            log.info('generating %s', dst)
        contents = subprocess.check_output(cmd).decode('utf-8')
        for hook in self.hooks:
            mod_name, attr_name = hook.split(':')
            mod = importlib.import_module(mod_name)
            hook_fn = getattr(mod, attr_name)
            contents = hook_fn(contents)
        with open(dst, 'w') as fp:
            fp.write(contents)

    def _build_resources(self, src):
        dst = os.path.join(
            os.path.dirname(os.path.dirname(src)),
            os.path.splitext(os.path.basename(src))[0]
        ) + '_rc.py'
        cmd = (
            sys.executable, '-m', 'PyQt5.pyrcc_main',
            src, '-o', dst,
        )
        if self.verbose:
            log.info('generating %s', dst)
        subprocess.check_call(cmd)

    def run(self):
        self.run_command('egg_info')
        ei_cmd = self.get_finalized_command('egg_info')
        for src in ei_cmd.filelist.files:
            if src.endswith('.qrc'):
                self._build_resources(src)
            if src.endswith('.ui'):
                self._build_ui(src)

# }}}

# Patched `build_py` command. {{{

class BuildPy(build_py):

    build_dependencies = []

    def run(self):
        for command in self.build_dependencies:
            self.run_command(command)
        build_py.run(self)

# }}}


def ensure_setup_requires(setuptools_spec, dependency_links=None, setup_requires=None):
    if 'PYTHONPATH' in os.environ:
        py_path = os.environ['PYTHONPATH'].split(os.pathsep)
    else:
        py_path = []
    # First, ensure the correct version of setuptools is active.
    setuptools_req = next(pkg_resources.parse_requirements('setuptools' + setuptools_spec))
    setuptools_dist = pkg_resources.get_distribution('setuptools')
    if setuptools_dist not in setuptools_req:
        setuptools_dist = setuptools.Distribution().fetch_build_eggs(str(setuptools_req))[0]
        py_path.insert(0, setuptools_dist.location)
        os.environ['PYTHONPATH'] = os.pathsep.join(py_path)
        args = [sys.executable] + sys.argv
        os.execv(args[0], args)
    # Second, install other setup requirements.
    setup_attrs = {}
    if dependency_links is not None:
        setup_attrs['dependency_links'] = dependency_links
    if setup_requires is not None:
        setup_attrs['setup_requires'] = setup_requires
    setup_dist = setuptools.Distribution(setup_attrs)
    setup_dist.parse_config_files(ignore_option_errors=True)
    if not setup_dist.setup_requires:
        return
    eggs_dir = setup_dist.get_egg_cache_dir()
    for dist in setup_dist.fetch_build_eggs(setup_dist.setup_requires):
        if dist.location.startswith(os.path.abspath(eggs_dir) + os.sep):
            py_path.insert(0, dist.location)
    os.environ['PYTHONPATH'] = os.pathsep.join(py_path)



#!/usr/bin/env python3

import argparse
import inspect
import os
import shutil
import subprocess
import sys
import traceback


WIN_DIR = os.path.dirname(os.path.abspath(__file__))
TOP_DIR = os.path.dirname(WIN_DIR)
NULL = open(os.devnull, 'r+b')
TEMP_DIR = r'C:\Temp'
PROG_DIR = r'C:\Progs'


sys.path.insert(0, TOP_DIR)
os.chdir(TOP_DIR)


if sys.stdout.isatty() and not sys.platform.startswith('win32'):

    def info(fmt, *args):
        s = fmt % args
        print('[34m' + s + '[0m')

else:

    def info(fmt, *args):
        s = fmt % args
        print(s)

if sys.stderr.isatty():

    def error(fmt, *args):
        s = fmt % args
        print('[31m' + s + '[0m', file=sys.stderr)

else:

    def error(fmt, *args):
        s = fmt % args
        print(s, file=sys.stderr)


class CommandExecutionException(Exception):

    def __init__(self, args, exitcode, stdout='', stderr=''):
        super().__init__()
        self.args = args
        self.exitcode = exitcode
        self.stdout = stdout
        self.stderr = stderr


class Environment:

    def __init__(self):
        self.dry_run = False
        self.verbose = False
        self._env = {}

    def setup(self):
        pass

    def get_realpath(self, path):
        return path

    def get_distdir(self):
        raise NotImplementedError

    def add_to_path(self, directory):
        raise NotImplementedError

    def run(self, args, cwd=None, redirection='auto'):
        if self.verbose:
            print('running', ' '.join(args))
        if self.dry_run:
            return 0, ''
        env = self._get_env()
        if 'auto' == redirection:
            to_pipe = not self.verbose
        elif 'never' == redirection:
            to_pipe = False
        elif 'always' == redirection:
            to_pipe = True
        else:
            raise ValueError('invalid redirection mode: %s' % redirection)
        if to_pipe:
            stdout = stderr = subprocess.PIPE
        else:
            stdout = stderr = None
        if cwd is None:
            old_cwd = None
        else:
            old_cwd = os.getcwd()
            os.chdir(cwd)
        try:
            proc = subprocess.Popen(args, env=env, stdout=stdout, stderr=stderr)
            stdout, stderr = proc.communicate()
            returncode = proc.wait()
        finally:
            if old_cwd is not None:
                os.chdir(old_cwd)
        if 0 != returncode:
            raise CommandExecutionException(args, returncode, stdout, stderr)
        return stdout

    def _get_env(self):
        env = dict(os.environ)
        env.update(self._env)
        return env


class WineEnvironment(Environment):

    def __init__(self, prefix):
        super().__init__()
        self._prefix = os.path.abspath(prefix)
        self._env['WINEPREFIX'] = self._prefix
        self._env['WINEARCH'] = 'win32'
        self._env['WINEDEBUG'] = '-all'
        self._env['WINETRICKS_OPT_SHAREDPREFIX'] = '1'

    def setup(self):
        if not os.path.exists(self._prefix):
            info('intializing Wine prefix')
            for cmd in (
                'env DISPLAY= wineboot --init',
                # Wait for previous command to finish.
                'wineserver -w',
                'winetricks --no-isolate --unattended corefonts vcrun2008',
                'winetricks win7',
            ):
                self.run(cmd.split())
        if self.dry_run:
            return
        tempdir = self.get_realpath(TEMP_DIR)
        progdir = self.get_realpath(PROG_DIR)
        distdir = self.get_distdir()
        for directory in (tempdir, progdir, distdir):
            if not os.path.exists(directory):
                os.makedirs(directory)

    def get_realpath(self, path):
        path = path.replace('\\', '/')
        if path.startswith('C:'):
            path = 'c:' + path[2:]
        return os.path.join(self._prefix, 'dosdevices', path)

    def get_distdir(self):
        return os.path.expanduser('~/.cache/wine')

    def add_to_path(self, directory):
        if self.verbose:
            print('adding "%s" to PATH' % directory)
        if self.dry_run:
            return
        path = self.run(('wine', 'cmd.exe', '/c', 'echo', '%PATH%'), redirection='always')
        path = path.decode().strip().replace("\\", "\\\\")
        directory = directory.replace("\\", "\\\\")
        regfile = r'%s\path.reg' % TEMP_DIR
        real_regfile = self.get_realpath(regfile)
        with open(real_regfile, 'w') as fp:
            fp.write('REGEDIT4\r\n')
            fp.write('\r\n')
            fp.write('[HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Session Manager\\Environment]\r\n')
            fp.write('"PATH"="%s;%s"\r\n' % (directory, path))
        try:
            self.run(('regedit', regfile))
        finally:
            os.unlink(real_regfile)

    def run(self, args, cwd=None, redirection='auto'):
        exe = args[0]
        if exe.endswith('.exe') or exe.endswith('.bat'):
            cmd = ('wine',) + tuple(args)
        else:
            cmd = args
        return super().run(cmd, cwd, redirection)

    def _get_env(self):
        env = super()._get_env()
        for var in (
            'PYTHONPATH',
            # Make sure GTK environment variables don't spill over to the Wine environment.
            'GDK_BACKEND',
            'GDK_NATIVE_WINDOWS',
            'GDK_PIXBUF_MODULE_FILE',
            'GDK_RENDERING',
            'GTK2_RC_FILES',
            'GTK3_MODULES',
            'GTK_DATA_PREFIX',
            'GTK_EXE_PREFIX',
            'GTK_IM_MODULE',
            'GTK_IM_MODULE_FILE',
            'GTK_MODULES',
            'GTK_PATH',
            'GTK_THEME',
        ):
            if var in env:
                del env[var]
        return env


class Win32Environment(Environment):

    def __init__(self):
        super().__init__()
        self._path = None

    def setup(self):
        if self.dry_run:
            return
        tempdir = self.get_realpath(TEMP_DIR)
        progdir = self.get_realpath(PROG_DIR)
        distdir = self.get_distdir()
        for directory in (tempdir, progdir, distdir):
            if not os.path.exists(directory):
                os.makedirs(directory)

    def get_distdir(self):
        return r'C:\Dist'

    def add_to_path(self, directory):
        if self.verbose:
            print('adding "%s" to PATH' % directory)
        if self._path is None:
            self._path = self.run(('cmd.exe', '/c', 'echo', '%PATH%'), redirection='always').strip().decode()
        self._path = directory + ';' + self._path
        batch = r'%s\setpath.bat' % TEMP_DIR
        with open(batch, 'w') as fp:
            fp.write('@ECHO off\r\n')
            fp.write('setx PATH "%s"\r\n' % self._path)
        try:
            self.run((batch,))
        finally:
            os.unlink(batch)


class Helper:

    DEPENDENCIES = (
        ('nsis', 'http://prdownloads.sourceforge.net/nsis/nsis-3.01-setup.exe',
         'f47f59956cdf1de42a5476649a4973bc11956987', None,
         # Note: the NSIS installer itself does support `/D` to set the
         # installation directory, and is installed in `C:\Program Files
         # (x86)\NSIS` on AppVeyor, but `C:\Program Files\NSIS` in Wine...
         ('/S',), (r'C:\Program Files\NSIS', r'C:\Program Files (x86)\NSIS')),
    )

    def __init__(self):
        self.dry_run = False
        self.verbose = False
        self.unattended = True
        self._parser = argparse.ArgumentParser(description='Windows Environment Helper')
        self._parser.add_argument('-v', '--verbose', dest='_verbose',
                                  action='store_true', help='enable verbose traces')
        self._parser.add_argument('-n', '--dry-run', dest='_dry_run',
                                  action='store_true', help='don\'t actuall run any command, but show what would be done')
        self._subparsers = self._parser.add_subparsers(dest='_command', metavar='COMMAND')

        for name in dir(self):
            if not name.startswith('cmd_'):
                continue
            cmd = getattr(self, name)
            argspec = inspect.getargspec(cmd)
            name = name[4:]
            doc = cmd.__doc__.strip().split('\n')
            help = doc[0]
            parser = self._subparsers.add_parser(name, help=help, add_help=False)
            args_help = {}
            for line in doc[1:]:
                line = line.strip()
                if not line:
                    continue
                name, doc = line.split(': ', 1)
                args_help[name] = doc
            short_opts = set()
            assert 'self' == argspec.args[0]
            for n, a in enumerate(argspec.args[1:]):
                if argspec.defaults is None or n >= len(argspec.defaults):
                    default = None
                else:
                    default = argspec.defaults[n]
                if default is None:
                    name_or_flags = (a,)
                    action = None
                else:
                    name = '--' + a
                    short_name = '-' + a[0]
                    if short_name not in short_opts:
                        short_opts.add(short_name)
                        name_or_flags = (short_name, name)
                    else:
                        name_or_flags = name,
                    if isinstance(default, bool):
                        action = 'store_' + str(not default).lower()
                    else:
                        action = 'store'
                parser.add_argument(*name_or_flags,
                                    action=action,
                                    default=default,
                                    help=args_help.get(a))
            if argspec.varargs is not None:
                parser.add_argument(argspec.varargs,
                                    nargs=argparse.REMAINDER,
                                    help=args_help.get(argspec.varargs))


        self._env = Environment()
        self._install_handlers = {
            'easy_install': self._easy_install,
            'pip'         : self._pip_install,
            '.exe'        : self._install_exe,
            '.msi'        : self._install_msi,
            '.rar'        : self._install_archive,
            '.zip'        : self._install_archive,
            '.whl'        : self._pip_install,
        }

    def _rmtree(self, path):
        if not os.path.exists(path):
            return
        if self.verbose:
            print('rm -rf', path)
        if self.dry_run:
            return
        shutil.rmtree(path)

    def _makedirs(self, path):
        if os.path.exists(path):
            return
        if self.verbose:
            print('mkdir -p', path)
        if self.dry_run:
            return
        os.makedirs(path)

    def _rename(self, old, new):
        if self.verbose:
            print('mv', old, new)
        if self.dry_run:
            return
        os.rename(old, new)

    def _copyfile(self, src, dst):
        if self.verbose:
            print('cp', src, dst)
        if self.dry_run:
            return
        shutil.copyfile(src, dst)

    def _copytree(self, src, dst):
        if self.verbose:
            print('cp -r', src, dst)
        if self.dry_run:
            return
        shutil.copytree(src, dst)

    def _extract(self, archive, destination, *args):
        if self.verbose:
            print('extracting %s to %s' % (archive, destination))
        cmd = ['7z.exe', 'x', archive, '-y', '-o%s' % destination]
        cmd.extend(args)
        realdir = self._env.get_realpath(destination)
        self._rmtree(realdir)
        self._env.run(cmd)
        if self.dry_run:
            return
        contents = os.listdir(realdir)
        if 1 == len(contents):
            tmpdir = '%s.tmp' % realdir
            os.rename(realdir, tmpdir)
            os.rename(os.path.join(tmpdir, contents[0]), realdir)
            os.rmdir(tmpdir)

    def _install_exe(self, filename, *options):
        cmd = [filename]
        cmd.extend(options)
        self._env.run(cmd)

    def _install_msi(self, filename, *options):
        cmd = ['msiexec.exe', '/i', filename]
        if self.unattended:
            cmd.append('/q')
        cmd.extend(options)
        self._env.run(cmd)

    def _install_archive(self, filename, install_basename, add_to_path=True):
        program_dir = r'%s\%s' % (PROG_DIR, install_basename)
        self._extract(filename, program_dir)
        if add_to_path:
            self._env.add_to_path(program_dir)

    def _easy_install(self, filename, *options):
        if not self.unattended and filename.endswith('.exe'):
            cmd = [filename]
        else:
            # Don't use easy_install.exe so setuptools
            # can update itself if needed.
            cmd = ['python.exe', '-m', 'easy_install']
            if not self.verbose:
                cmd.append('--quiet')
            cmd.extend(options)
            cmd.extend(('--', filename))
        self._env.run(cmd)

    def _pip_install(self, *args):
        # Don't use pip.exe so pip can update itself if needed.
        cmd = ['python.exe', '-m', 'pip',
               '--timeout=5',
               '--retries=2',
               '--disable-pip-version-check']
        if not self.verbose:
            cmd.append('--quiet')
        cmd.extend(('install', '--upgrade'))
        cmd.extend(args)
        self._env.run(cmd)

    def _download(self, url, checksum):
        from plover_build_utils.download import DOWNLOADS_DIR, download
        if not self.dry_run:
            return download(url, checksum)
        return os.path.join(DOWNLOADS_DIR, os.path.basename(url))

    def install(self, name, src, checksum, handler_format=None, handler_args=(), path_dir=None):
        info('installing %s', name)
        if src.startswith('pip:'):
            handler_format = 'pip'
            dst = src[4:]
        else:
            distdir = self._env.get_distdir()
            dst = os.path.join(distdir, os.path.basename(src))
            dst = self._download(src, checksum)
            if handler_format is None:
                root, handler_format = os.path.splitext(dst)
        handler = self._install_handlers.get(handler_format)
        assert handler, 'no handler for installing format %s' % handler_format
        handler(dst, *handler_args)
        if path_dir is not None:
            if isinstance(path_dir, (list, tuple)):
                for d in path_dir:
                    self._env.add_to_path(d)
            else:
                self._env.add_to_path(path_dir)

    def cmd_help(self, *commands):
        '''print detailed help'''
        if not commands:
            commands = self._subparsers.choices.keys()
            self._parser.print_help()
        for name in commands:
            print()
            parser = self._subparsers.choices.get(name)
            if parser is None:
                raise ValueError('unknown command name: %s' % name)
            print('%s COMMAND' % name.upper())
            print()
            parser.print_help()

    def cmd_setup(self, interactive=False):
        '''setup environment

        interactive: don't run unattended
        '''
        self.unattended = not interactive
        self._env.setup()
        for name, src, checksum, handler_format, handler_args, path_dir in self.DEPENDENCIES:
            self.install(name, src, checksum, handler_format=handler_format, handler_args=handler_args, path_dir=path_dir)
        info('install requirements')
        self._env.run(('python.exe', '-m', 'plover_build_utils.get_pip', '--upgrade'))
        self._env.run(('python.exe', '-m', 'plover_build_utils.install_wheels', '-r', 'requirements.txt'))

    def cmd_run(self, executable, *args):
        '''run command in environment

        executable: name of the executable to run
        args: additional arguments to pass to the executable
        '''
        self._env.run((executable,) + tuple(args), redirection='never')

    def cmd_dist(self, trim=False, zipdir=False, installer=False):
        '''create windows distribution

        trim: trim the result to reduce size
        zipdir: zip the resulting directory
        installer: also create an installer
        '''
        self._rmtree('build')
        self._rmtree('dist')
        info('creating distribution')
        cmd = ['python.exe', 'setup.py']
        if self.verbose:
            cmd.append('-v')
        cmd.append('bdist_win')
        if trim:
            cmd.append('-t')
        if zipdir:
            cmd.append('-z')
        if installer:
            cmd.append('-i')
        self._env.run(cmd)

    def main(self, args):
        opts = self._parser.parse_args(args)
        self.dry_run = self._env.dry_run = opts._dry_run
        self.verbose = self._env.verbose = opts._verbose
        args = {
            name: value
            for name, value
            in opts._get_kwargs()
            if not name.startswith('_')
        }
        cmd = getattr(self, 'cmd_%s' % opts._command)
        argspec = inspect.getargspec(cmd)
        args = []
        assert 'self' == argspec.args[0]
        for a in argspec.args[1:]:
            args.append(getattr(opts, a))
        if argspec.varargs is not None:
            args.extend(getattr(opts, argspec.varargs))
        try:
            cmd(*args)
        except CommandExecutionException as e:
            if e.stdout:
                print(e.stdout.decode(), file=sys.stdout)
            if e.stderr:
                print(e.stderr.decode(), file=sys.stderr)
            error('execution failed, returned %u: %s',
                  e.exitcode, ' '.join(e.args))
            return e.exitcode
        except Exception as e:
            error('exception: %s', str(e))
            traceback.print_exc()
            return -1
        return 0


class WineHelper(Helper):

    DEPENDENCIES = (
        ('Python', 'https://www.python.org/ftp/python/3.6.2/python-3.6.2.exe',
         'cd9744b142eca832f9534390676e6cfb84bf655d', None, ('PrependPath=1', '/S'), None),
    ) + Helper.DEPENDENCIES

    def __init__(self):
        super().__init__()
        self._env = WineEnvironment(os.path.join(WIN_DIR, '.wine'))

    def cmd_run(self, executable, *args):
        '''run command in environment

        executable: executable to run (the executable will be automatically run with Wine when it ends with ".exe")
        args: additional arguments to pass to the executable
        '''
        super().cmd_run(executable, *args)


class Win32Helper(Helper):

    def __init__(self):
        super().__init__()
        self._env = Win32Environment()


if '__main__' == __name__:
    if sys.platform.startswith('linux'):
        h = WineHelper()
    elif sys.platform.startswith('win32'):
        h = Win32Helper()
    sys.exit(h.main(sys.argv[1:]))





import os

import pkg_resources


ASSET_SCHEME = 'asset:'

def _asset_split(resource_name):
    components = resource_name[len(ASSET_SCHEME):].split(':', 1)
    if len(components) != 2:
        raise ValueError('invalid asset: %s' % resource_name)
    return components

def resource_exists(resource_name):
    if resource_name.startswith(ASSET_SCHEME):
        return pkg_resources.resource_exists(*_asset_split(resource_name))
    return os.path.exists(resource_name)

def resource_filename(resource_name):
    if resource_name.startswith(ASSET_SCHEME):
        return pkg_resources.resource_filename(*_asset_split(resource_name))
    return resource_name

def resource_timestamp(resource_name):
    filename = resource_filename(resource_name)
    return os.path.getmtime(filename)




from functools import partial
import json

from PyQt5.QtCore import QCoreApplication, Qt
from PyQt5.QtGui import QCursor, QIcon, QKeySequence
from PyQt5.QtWidgets import (
    QMainWindow,
    QMenu,
)

from plover import log
from plover.oslayer import wmctrl
from plover.registry import registry
from plover.resource import resource_filename

from plover.gui_qt.log_qt import NotificationHandler
from plover.gui_qt.main_window_ui import Ui_MainWindow
from plover.gui_qt.config_window import ConfigWindow
from plover.gui_qt.about_dialog import AboutDialog
from plover.gui_qt.trayicon import TrayIcon
from plover.gui_qt.utils import WindowState, find_menu_actions


class MainWindow(QMainWindow, Ui_MainWindow, WindowState):

    ROLE = 'main'

    def __init__(self, engine, use_qt_notifications):
        super().__init__()
        self.setupUi(self)
        if hasattr(self, 'setUnifiedTitleAndToolBarOnMac'):
            self.setUnifiedTitleAndToolBarOnMac(True)
        self._engine = engine
        self._active_dialogs = {}
        self._dialog_class = {
            'about'             : AboutDialog,
            'configuration'     : ConfigWindow,
        }
        all_actions = find_menu_actions(self.menubar)
        # Dictionaries.
        self.dictionaries = self.scroll_area.widget()
        self.dictionaries.add_translation.connect(self._add_translation)
        self.dictionaries.setFocus()
        edit_menu = all_actions['menu_Edit'].menu()
        edit_menu.addAction(self.dictionaries.action_Undo)
        edit_menu.addSeparator()
        edit_menu.addMenu(self.dictionaries.menu_AddDictionaries)
        edit_menu.addAction(self.dictionaries.action_EditDictionaries)
        edit_menu.addAction(self.dictionaries.action_RemoveDictionaries)
        edit_menu.addSeparator()
        edit_menu.addAction(self.dictionaries.action_MoveDictionariesUp)
        edit_menu.addAction(self.dictionaries.action_MoveDictionariesDown)
        # Tray icon.
        self._trayicon = TrayIcon()
        self._trayicon.enable()
        self._trayicon.clicked.connect(self._engine.toggle_output)
        if use_qt_notifications:
            handler = NotificationHandler()
            handler.emitSignal.connect(self._trayicon.log)
            log.add_handler(handler)
        popup_menu = QMenu()
        for action_name in (
            'action_ToggleOutput',
            'action_Reconnect',
            '',
            'menu_Tools',
            '',
            'action_Configure',
            '',
            'menu_Help',
            '',
            'action_Show',
            'action_Quit',
        ):
            if action_name:
                popup_menu.addAction(all_actions[action_name])
            else:
                popup_menu.addSeparator()
        self._trayicon.set_menu(popup_menu)
        engine.signal_connect('machine_state_changed', self._trayicon.update_machine_state)
        engine.signal_connect('quit', self.on_quit)
        self.action_Quit.triggered.connect(engine.quit)
        # Populate tools bar/menu.
        tools_menu = all_actions['menu_Tools'].menu()
        # Toolbar popup menu for selecting which tools are shown.
        self.toolbar_menu = QMenu()
        self.toolbar.setContextMenuPolicy(Qt.CustomContextMenu)
        self.toolbar.customContextMenuRequested.connect(
            lambda: self.toolbar_menu.popup(QCursor.pos())
        )
        for tool_plugin in registry.list_plugins('gui.qt.tool'):
            tool = tool_plugin.obj
            action_parameters = []
            if tool.ICON is not None:
                icon = tool.ICON
                # Internal QT resources start with a `:`.
                if not icon.startswith(':'):
                    icon = resource_filename(icon)
                action_parameters.append(QIcon(icon))
            action_parameters.append(tool.TITLE)
            toolbar_action = None
            for parent in (tools_menu, self.toolbar, self.toolbar_menu):
                action = parent.addAction(*action_parameters)
                action.setObjectName(tool_plugin.name)
                if tool.__doc__ is not None:
                    action.setToolTip(tool.__doc__)
                if tool.SHORTCUT is not None:
                    action.setShortcut(QKeySequence.fromString(tool.SHORTCUT))
                if parent == self.toolbar_menu:
                    action.setCheckable(True)
                    action.setChecked(True)
                    assert toolbar_action is not None
                    action.toggled.connect(toolbar_action.setVisible)
                else:
                    if parent == self.toolbar:
                        toolbar_action = action
                    action.triggered.connect(partial(self._activate_dialog,
                                                     tool_plugin.name,
                                                     args=()))
            self._dialog_class[tool_plugin.name] = tool
        engine.signal_connect('output_changed', self.on_output_changed)
        # Machine.
        self.machine_type.addItems(
            _(plugin.name)
            for plugin in registry.list_plugins('machine')
        )
        engine.signal_connect('config_changed', self.on_config_changed)
        engine.signal_connect('machine_state_changed',
            lambda machine, state:
            self.machine_state.setText(_(state.capitalize()))
        )
        self.restore_state()
        # Commands.
        engine.signal_connect('add_translation', partial(self._add_translation, manage_windows=True))
        engine.signal_connect('focus', self._focus)
        engine.signal_connect('configure', partial(self._configure, manage_windows=True))
        engine.signal_connect('lookup', partial(self._activate_dialog, 'lookup',
                                                manage_windows=True))
        # Load the configuration (but do not start the engine yet).
        if not engine.load_config():
            self.on_configure()
        # Apply configuration settings.
        config = self._engine.config
        self.machine_type.setCurrentText(config['machine_type'])
        self._configured = False
        self.dictionaries.on_config_changed(config)
        self.set_visible(not config['start_minimized'])
        # Start the engine.
        engine.start()

    def set_visible(self, visible):
        if visible:
            self.show()
        else:
            if self._trayicon.is_enabled():
                self.hide()
            else:
                self.showMinimized()

    def _activate_dialog(self, name, args=(), manage_windows=False):
        if manage_windows:
            previous_window = wmctrl.GetForegroundWindow()
        dialog = self._active_dialogs.get(name)
        if dialog is None:
            dialog_class = self._dialog_class[name]
            dialog = self._active_dialogs[name] = dialog_class(self._engine, *args)
            dialog.setWindowIcon(self.windowIcon())
            def on_finished():
                del self._active_dialogs[name]
                dialog.deleteLater()
                if manage_windows and previous_window is not None:
                    wmctrl.SetForegroundWindow(previous_window)
            dialog.finished.connect(on_finished)
        dialog.show()
        dialog.activateWindow()
        dialog.raise_()

    def _add_translation(self, dictionary=None, manage_windows=False):
        if not dictionary:
            dictionary = None
        self._activate_dialog('add_translation', args=(dictionary,),
                              manage_windows=manage_windows)

    def _focus(self):
        self.set_visible(True)
        self.activateWindow()
        self.raise_()

    def _configure(self, manage_windows=False):
        self._activate_dialog('configuration', manage_windows=manage_windows)

    def _lookup(self, manage_windows=False):
        self._activate_dialog('lookup', manage_windows=manage_windows)

    def _restore_state(self, settings):
        if settings.contains('hidden_toolbar_tools'):
            hidden_toolbar_tools = json.loads(settings.value('hidden_toolbar_tools'))
            for action in self.toolbar_menu.actions():
                if action.objectName() in hidden_toolbar_tools:
                    action.setChecked(False)

    def _save_state(self, settings):
        hidden_toolbar_tools = set()
        for action in self.toolbar_menu.actions():
            if not action.isChecked():
                hidden_toolbar_tools.add(action.objectName())
        settings.setValue('hidden_toolbar_tools', json.dumps(list(sorted(hidden_toolbar_tools))))

    def on_config_changed(self, config_update):
        if 'machine_type' in config_update:
            self.machine_type.setCurrentText(config_update['machine_type'])
        if not self._configured:
            self._configured = True
            if config_update.get('show_suggestions_display', False):
                self._activate_dialog('suggestions')
            if config_update.get('show_stroke_display', False):
                self._activate_dialog('paper_tape')

    def on_machine_changed(self, machine_type):
        self._engine.config = { 'machine_type': machine_type }

    def on_output_changed(self, enabled):
        self._trayicon.update_output(enabled)
        self.output_enable.setChecked(enabled)
        self.output_disable.setChecked(not enabled)
        self.action_ToggleOutput.setChecked(enabled)

    def on_toggle_output(self, enabled):
        self._engine.output = enabled

    def on_enable_output(self):
        self.on_toggle_output(True)

    def on_disable_output(self):
        self.on_toggle_output(False)

    def on_configure(self):
        self._configure()

    def on_reconnect(self):
        self._engine.reset_machine()

    def on_manage_dictionaries(self):
        self._activate_dialog('dictionary_manager')

    def on_about(self):
        self._activate_dialog('about')

    def on_quit(self):
        for dialog in list(self._active_dialogs.values()):
            dialog.close()
        self.save_state()
        self._trayicon.disable()
        self.hide()
        QCoreApplication.quit()

    def on_show(self):
        self._focus()

    def closeEvent(self, event):
        self.hide()
        if not self._trayicon.is_enabled():
            self._engine.quit()
        event.ignore()




import pytest

from plover.machine.keymap import Keymap


def new_keymap():
    return Keymap(('k%u' % n for n in range(8)),
                  ('a%u' % n for n in range(4)))

BINDINGS_LIST = (
    ('k0', 'a0'),
    ('k1', 'a3'),
    ('k4', 'a1'),
    ('k5', 'a2'),
    ('k7', 'a1'),
)
BINDINGS_DICT = dict(BINDINGS_LIST)
MAPPINGS_LIST = (
    ('a0', ('k0',)),
    ('a1', ('k4', 'k7')),
    ('a2', ('k5',)),
    ('a3', ('k1',)),
)
MAPPINGS_DICT = dict(MAPPINGS_LIST)
MAPPINGS_FULL = dict(MAPPINGS_DICT)
MAPPINGS_FULL.update({'no-op': ()})


def test_keymap_init():
    k = new_keymap()
    assert k.get_bindings() == {}
    assert k.get_mappings() == {}

def test_keymap_set_bindings():
    # Set bindings from a dictionary.
    k = new_keymap()
    k.set_bindings(BINDINGS_DICT)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL
    # Set bindings from a list of tuples.
    k = new_keymap()
    k.set_bindings(BINDINGS_LIST)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL

def test_keymap_set_mappings():
    # Set mappings from a dictionary.
    k = new_keymap()
    k.set_mappings(MAPPINGS_DICT)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL
    # Set mappings from a list of tuples.
    k = new_keymap()
    k.set_mappings(MAPPINGS_LIST)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL

def test_keymap_setitem():
    bindings = dict(BINDINGS_DICT)
    mappings = dict(MAPPINGS_FULL)
    k = new_keymap()
    k.set_mappings(mappings)
    # Bind to one key.
    k['a3'] = 'k6'
    del bindings['k1']
    bindings['k6'] = 'a3'
    mappings['a3'] = ('k6',)
    assert k.get_bindings() == bindings
    assert k.get_mappings() == mappings
    # Bind to multiple keys.
    k['a3'] = ('k6', 'k1')
    bindings['k1'] = 'a3'
    bindings['k6'] = 'a3'
    mappings['a3'] = ('k1', 'k6',)
    assert k.get_bindings() == bindings
    assert k.get_mappings() == mappings
    # If a key is already mapped (k0), don't override it.
    k['a3'] = ('k0', 'k2')
    del bindings['k1']
    del bindings['k6']
    bindings['k2'] = 'a3'
    mappings['a3'] = ('k2',)
    assert k.get_bindings() == bindings
    assert k.get_mappings() == mappings
    # Assert on invalid action.
    with pytest.raises(AssertionError):
        k['a9'] = 'k0'



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for steno.py."""

from plover.steno import normalize_steno, Stroke

from . import parametrize


NORMALIZE_TESTS = (
    # TODO: More cases
    lambda: ('S', 'S'),
    lambda: ('S-', 'S'),
    lambda: ('-S', '-S'),
    lambda: ('ES', 'ES'),
    lambda: ('-ES', 'ES'),
    lambda: ('TW-EPBL', 'TWEPBL'),
    lambda: ('TWEPBL', 'TWEPBL'),
    lambda: ('19', '1-9'),
    lambda: ('14', '14'),
    lambda: ('146', '14-6'),
    lambda: ('67', '-67'),
    lambda: ('6', '-6'),
    lambda: ('9', '-9'),
    lambda: ('5', '5'),
    lambda: ('0', '0'),
    lambda: ('456', '456'),
    lambda: ('46', '4-6'),
    lambda: ('4*6', '4*6'),
    lambda: ('456', '456'),
    lambda: ('S46', 'S4-6'),
    # Number key.
    lambda: ('#S', '#S'),
    lambda: ('#A', '#A'),
    lambda: ('#0', '0'),
    lambda: ('#6', '-6'),
    # Implicit hyphens.
    lambda: ('SA-', 'SA'),
    lambda: ('SA-R', 'SAR'),
    lambda: ('-O', 'O'),
    lambda: ('S*-R', 'S*R'),
    lambda: ('S-*R', 'S*R'),
)

@parametrize(NORMALIZE_TESTS)
def test_normalize_steno(steno, strokes):
    result = '/'.join(normalize_steno(steno))
    msg = 'normalize_steno(%r)=%r != %r' % (
        steno, result, strokes,
    )
    assert result == strokes, msg


STROKE_TESTS = (
    lambda: (['S-'], ['S-'], 'S'),
    lambda: (['S-', 'T-'], ['S-', 'T-'], 'ST'),
    lambda: (['T-', 'S-'], ['S-', 'T-'], 'ST'),
    lambda: (['-P', '-P'], ['-P'], '-P'),
    lambda: (['-P', 'X-'], ['X-', '-P'], 'X-P'),
    lambda: (['#', 'S-', '-T'], ['1-', '-9'], '1-9'),
)

@parametrize(STROKE_TESTS)
def test_stroke(keys, steno_keys, rtfcre):
    stroke = Stroke(keys)
    assert stroke.steno_keys == steno_keys
    assert stroke.rtfcre == rtfcre



# -*- coding: utf-8 -*-

import re


# Mapping of "standard" keynames (derived from X11 keysym names) to Unicode.
KEYNAME_TO_CHAR = {
    # Generated using:
    #
    # from Xlib import XK
    # from plover.oslayer.xkeyboardcontrol import keysym_to_string
    # for kn, ks in sorted({
    #     name[3:].lower(): getattr(XK, name)
    #     for name in sorted(dir(XK))
    #     if name.startswith('XK_')
    # }.items()):
    #     us = keysym_to_string(ks)
    #     if us == kn or not us:
    #         continue
    # print '    %-20r: %8r, # %s' % (kn, us, us)
    'aacute'            :  '\xe1', # á
    'acircumflex'       :  '\xe2', # â
    'acute'             :  '\xb4', # ´
    'adiaeresis'        :  '\xe4', # ä
    'ae'                :  '\xe6', # æ
    'agrave'            :  '\xe0', # à
    'ampersand'         :     '&', # &
    'apostrophe'        :     "'", # '
    'aring'             :  '\xe5', # å
    'asciicircum'       :     '^', # ^
    'asciitilde'        :     '~', # ~
    'asterisk'          :     '*', # *
    'at'                :     '@', # @
    'atilde'            :  '\xe3', # ã
    'backslash'         :    '\\', # \
    'bar'               :     '|', # |
    'braceleft'         :     '{', # {
    'braceright'        :     '}', # }
    'bracketleft'       :     '[', # [
    'bracketright'      :     ']', # ]
    'brokenbar'         :  '\xa6', # ¦
    'ccedilla'          :  '\xe7', # ç
    'cedilla'           :  '\xb8', # ¸
    'cent'              :  '\xa2', # ¢
    'clear'             :  '\x0b', # 
    'colon'             :     ':', # :
    'comma'             :     ',', # ,
    'copyright'         :  '\xa9', # ©
    'currency'          :  '\xa4', # ¤
    'degree'            :  '\xb0', # °
    'diaeresis'         :  '\xa8', # ¨
    'division'          :  '\xf7', # ÷
    'dollar'            :     '$', # $
    'eacute'            :  '\xe9', # é
    'ecircumflex'       :  '\xea', # ê
    'ediaeresis'        :  '\xeb', # ë
    'egrave'            :  '\xe8', # è
    'equal'             :     '=', # =
    'eth'               :  '\xf0', # ð
    'exclam'            :     '!', # !
    'exclamdown'        :  '\xa1', # ¡
    'grave'             :     '`', # `
    'greater'           :     '>', # >
    'guillemotleft'     :  '\xab', # «
    'guillemotright'    :  '\xbb', # »
    'hyphen'            :  '\xad', # ­
    'iacute'            :  '\xed', # í
    'icircumflex'       :  '\xee', # î
    'idiaeresis'        :  '\xef', # ï
    'igrave'            :  '\xec', # ì
    'less'              :     '<', # <
    'macron'            :  '\xaf', # ¯
    'masculine'         :  '\xba', # º
    'minus'             :     '-', # -
    'mu'                :  '\xb5', # µ
    'multiply'          :  '\xd7', # ×
    'nobreakspace'      :  '\xa0', #  
    'notsign'           :  '\xac', # ¬
    'ntilde'            :  '\xf1', # ñ
    'numbersign'        :     '#', # #
    'oacute'            :  '\xf3', # ó
    'ocircumflex'       :  '\xf4', # ô
    'odiaeresis'        :  '\xf6', # ö
    'ograve'            :  '\xf2', # ò
    'onehalf'           :  '\xbd', # ½
    'onequarter'        :  '\xbc', # ¼
    'onesuperior'       :  '\xb9', # ¹
    'ooblique'          :  '\xd8', # Ø
    'ordfeminine'       :  '\xaa', # ª
    'oslash'            :  '\xf8', # ø
    'otilde'            :  '\xf5', # õ
    'paragraph'         :  '\xb6', # ¶
    'parenleft'         :     '(', # (
    'parenright'        :     ')', # )
    'percent'           :     '%', # %
    'period'            :     '.', # .
    'periodcentered'    :  '\xb7', # ·
    'plus'              :     '+', # +
    'plusminus'         :  '\xb1', # ±
    'question'          :     '?', # ?
    'questiondown'      :  '\xbf', # ¿
    'quotedbl'          :     '"', # "
    'quoteleft'         :     '`', # `
    'quoteright'        :     "'", # '
    'registered'        :  '\xae', # ®
    'return'            :    '\r', # 
    'section'           :  '\xa7', # §
    'semicolon'         :     ';', # ;
    'slash'             :     '/', # /
    'space'             :     ' ', #  
    'ssharp'            :  '\xdf', # ß
    'sterling'          :  '\xa3', # £
    'tab'               :    '\t', # 	
    'thorn'             :  '\xfe', # þ
    'threequarters'     :  '\xbe', # ¾
    'threesuperior'     :  '\xb3', # ³
    'twosuperior'       :  '\xb2', # ²
    'uacute'            :  '\xfa', # ú
    'ucircumflex'       :  '\xfb', # û
    'udiaeresis'        :  '\xfc', # ü
    'ugrave'            :  '\xf9', # ù
    'underscore'        :     '_', # _
    'yacute'            :  '\xfd', # ý
    'ydiaeresis'        :  '\xff', # ÿ
    'yen'               :  '\xa5', # ¥
}
for char in (
    '0123456789'
    'abcdefghijklmnopqrstuvwxyz'
):
    KEYNAME_TO_CHAR[char] = char
CHAR_TO_KEYNAME = {
    char: name
    for name, char in KEYNAME_TO_CHAR.items()
}


_SPLIT_RX = re.compile(r'(\s+|(?:\w+(?:\s*\()?)|.)')

def parse_key_combo(combo_string, key_name_to_key_code=None):

    if key_name_to_key_code is None:
        key_name_to_key_code = lambda key_name: key_name

    key_events = []
    down_keys = []
    token = None
    count = 0

    def _raise_error(exception, details):
        msg = '%s in "%s"' % (
            details,
            combo_string[:count] +
            '[' + token + ']' +
            combo_string[count+len(token):],
        )
        raise exception(msg)

    for token in _SPLIT_RX.split(combo_string):
        if not token:
            continue

        if token.isspace():
            pass

        elif re.match(r'\w', token):

            if token.endswith('('):
                key_name = token[:-1].rstrip().lower()
                release = False
            else:
                key_name = token.lower()
                release = True

            key_code = key_name_to_key_code(key_name)
            if key_code is None:
                _raise_error(ValueError, 'unknown key')
            elif key_code in down_keys:
                _raise_error(ValueError, 'key "%s" already pressed' % key_name)

            key_events.append((key_code, True))

            if release:
                key_events.append((key_code, False))
            else:
                down_keys.append(key_code)

        elif token == ')':
            if not down_keys:
                _raise_error(SyntaxError, 'unbalanced ")"')
            key_code = down_keys.pop()
            key_events.append((key_code, False))

        else:
            _raise_error(SyntaxError, 'invalid character "%s"' % token)

        count += len(token)

    if down_keys:
        _raise_error(SyntaxError, 'unbalanced "("')

    return key_events


def add_modifiers_aliases(dictionary):
    ''' Add aliases for common modifiers to a dictionary of key name to key code.

    - add `mod` for `mod_l` aliases for `alt`, `control`, `shift` and `super`
    - add `command` and `windows` aliases for `super`
    - add `option` alias for `alt`
    '''
    for name, extra_aliases in (
        ('control', ''               ),
        ('shift'  , ''               ),
        ('super'  , 'command windows'),
        ('alt'    , 'option'         ,)
    ):
        code = dictionary[name + '_l']
        dictionary[name] = code
        for alias in extra_aliases.split():
            dictionary[alias] = code



# -*- coding: utf-8 -*-

import sys
import codecs
import ctypes
from ctypes import windll, wintypes
from collections import defaultdict, namedtuple

from plover.key_combo import CHAR_TO_KEYNAME, add_modifiers_aliases
from plover.misc import popcount_8


GetKeyboardLayout = windll.user32.GetKeyboardLayout
GetWindowThreadProcessId = windll.user32.GetWindowThreadProcessId
MapVirtualKeyEx = windll.user32.MapVirtualKeyExW
ToUnicodeEx = windll.user32.ToUnicodeEx


def enum(name, items):
    ''' Helper to simulate an Enum. '''
    keys, values = zip(*items)
    t = namedtuple(name, keys)
    return t._make(values)

# Shift state enum. {{{

SHIFT_STATE = enum('SHIFT_STATE', ((mod, n) for n, mod in enumerate(
    'BASE SHIFT CTRL SHIFT_CTRL MENU SHIFT_MENU MENU_CTRL SHIFT_MENU_CTRL'.split()
)))

def shift_state_str(ss):
    s = ''
    for mod_state, mod_str in (
        (SHIFT_STATE.SHIFT, 'SHIFT'),
        (SHIFT_STATE.CTRL,  'CTRL'),
        (SHIFT_STATE.MENU,  'MENU'),
    ):
        if (ss & mod_state) != 0:
            s += mod_str + '+'
    return s

# }}}

# Virtual keys enum. {{{

vk_dict = {
    'LBUTTON'            : 0x01,
    'RBUTTON'            : 0x02,
    'CANCEL'             : 0x03,
    'MBUTTON'            : 0x04,
    'XBUTTON1'           : 0x05,
    'XBUTTON2'           : 0x06,
    'BACK'               : 0x08,
    'TAB'                : 0x09,
    'CLEAR'              : 0x0C,
    'RETURN'             : 0x0D,
    'SHIFT'              : 0x10,
    'CONTROL'            : 0x11,
    'MENU'               : 0x12,
    'PAUSE'              : 0x13,
    'CAPITAL'            : 0x14,
    'KANA'               : 0x15,
    'JUNJA'              : 0x17,
    'FINAL'              : 0x18,
    'HANJA'              : 0x19,
    'ESCAPE'             : 0x1B,
    'CONVERT'            : 0x1C,
    'NONCONVERT'         : 0x1D,
    'ACCEPT'             : 0x1E,
    'MODECHANGE'         : 0x1F,
    'SPACE'              : 0x20,
    'PRIOR'              : 0x21,
    'NEXT'               : 0x22,
    'END'                : 0x23,
    'HOME'               : 0x24,
    'LEFT'               : 0x25,
    'UP'                 : 0x26,
    'RIGHT'              : 0x27,
    'DOWN'               : 0x28,
    'SELECT'             : 0x29,
    'PRINT'              : 0x2A,
    'EXECUTE'            : 0x2B,
    'SNAPSHOT'           : 0x2C,
    'INSERT'             : 0x2D,
    'DELETE'             : 0x2E,
    'HELP'               : 0x2F,
    'LWIN'               : 0x5B,
    'RWIN'               : 0x5C,
    'APPS'               : 0x5D,
    'SLEEP'              : 0x5F,
    'NUMPAD0'            : 0x60,
    'NUMPAD1'            : 0x61,
    'NUMPAD2'            : 0x62,
    'NUMPAD3'            : 0x63,
    'NUMPAD4'            : 0x64,
    'NUMPAD5'            : 0x65,
    'NUMPAD6'            : 0x66,
    'NUMPAD7'            : 0x67,
    'NUMPAD8'            : 0x68,
    'NUMPAD9'            : 0x69,
    'MULTIPLY'           : 0x6A,
    'ADD'                : 0x6B,
    'SEPARATOR'          : 0x6C,
    'SUBTRACT'           : 0x6D,
    'DECIMAL'            : 0x6E,
    'DIVIDE'             : 0x6F,
    'F1'                 : 0x70,
    'F2'                 : 0x71,
    'F3'                 : 0x72,
    'F4'                 : 0x73,
    'F5'                 : 0x74,
    'F6'                 : 0x75,
    'F7'                 : 0x76,
    'F8'                 : 0x77,
    'F9'                 : 0x78,
    'F10'                : 0x79,
    'F11'                : 0x7A,
    'F12'                : 0x7B,
    'F13'                : 0x7C,
    'F14'                : 0x7D,
    'F15'                : 0x7E,
    'F16'                : 0x7F,
    'F17'                : 0x80,
    'F18'                : 0x81,
    'F19'                : 0x82,
    'F20'                : 0x83,
    'F21'                : 0x84,
    'F22'                : 0x85,
    'F23'                : 0x86,
    'F24'                : 0x87,
    'NUMLOCK'            : 0x90,
    'SCROLL'             : 0x91,
    'OEM_NEC_EQUAL'      : 0x92,
    'OEM_FJ_JISHO'       : 0x92,
    'OEM_FJ_MASSHOU'     : 0x93,
    'OEM_FJ_TOUROKU'     : 0x94,
    'OEM_FJ_LOYA'        : 0x95,
    'OEM_FJ_ROYA'        : 0x96,
    'LSHIFT'             : 0xA0,
    'RSHIFT'             : 0xA1,
    'LCONTROL'           : 0xA2,
    'RCONTROL'           : 0xA3,
    'LMENU'              : 0xA4,
    'RMENU'              : 0xA5,
    'BROWSER_BACK'       : 0xA6,
    'BROWSER_FORWARD'    : 0xA7,
    'BROWSER_REFRESH'    : 0xA8,
    'BROWSER_STOP'       : 0xA9,
    'BROWSER_SEARCH'     : 0xAA,
    'BROWSER_FAVORITES'  : 0xAB,
    'BROWSER_HOME'       : 0xAC,
    'VOLUME_MUTE'        : 0xAD,
    'VOLUME_DOWN'        : 0xAE,
    'VOLUME_UP'          : 0xAF,
    'MEDIA_NEXT_TRACK'   : 0xB0,
    'MEDIA_PREV_TRACK'   : 0xB1,
    'MEDIA_STOP'         : 0xB2,
    'MEDIA_PLAY_PAUSE'   : 0xB3,
    'LAUNCH_MAIL'        : 0xB4,
    'LAUNCH_MEDIA_SELECT': 0xB5,
    'LAUNCH_APP1'        : 0xB6,
    'LAUNCH_APP2'        : 0xB7,
    'OEM_1'              : 0xBA,
    'OEM_PLUS'           : 0xBB,
    'OEM_COMMA'          : 0xBC,
    'OEM_MINUS'          : 0xBD,
    'OEM_PERIOD'         : 0xBE,
    'OEM_2'              : 0xBF,
    'OEM_3'              : 0xC0,
    'OEM_4'              : 0xDB,
    'OEM_5'              : 0xDC,
    'OEM_6'              : 0xDD,
    'OEM_7'              : 0xDE,
    'OEM_8'              : 0xDF,
    'OEM_AX'             : 0xE1,
    'OEM_102'            : 0xE2,
    'ICO_HELP'           : 0xE3,
    'ICO_00'             : 0xE4,
    'PROCESSKEY'         : 0xE5,
    'ICO_CLEAR'          : 0xE6,
    'PACKET'             : 0xE7,
    'OEM_RESET'          : 0xE9,
    'OEM_JUMP'           : 0xEA,
    'OEM_PA1'            : 0xEB,
    'OEM_PA2'            : 0xEC,
    'OEM_PA3'            : 0xED,
    'OEM_WSCTRL'         : 0xEE,
    'OEM_CUSEL'          : 0xEF,
    'OEM_ATTN'           : 0xF0,
    'OEM_FINISH'         : 0xF1,
    'OEM_COPY'           : 0xF2,
    'OEM_AUTO'           : 0xF3,
    'OEM_ENLW'           : 0xF4,
    'OEM_BACKTAB'        : 0xF5,
    'ATTN'               : 0xF6,
    'CRSEL'              : 0xF7,
    'EXSEL'              : 0xF8,
    'EREOF'              : 0xF9,
    'PLAY'               : 0xFA,
    'ZOOM'               : 0xFB,
    'NONAME'             : 0xFC,
    'PA1'                : 0xFD,
    'OEM_CLEAR'          : 0xFE,
}
vk_dict['HANGEUL'] = vk_dict['KANA']
vk_dict['HANGUL']  = vk_dict['KANA']
vk_dict['KANJI']   = vk_dict['HANJA']
for digit in range(10):
    vk_dict['DIGIT%u' % digit] = 0x30 + digit
for anum in range(26):
    vk_dict[chr(ord('A')+anum)] = 0x41 + anum
VK = enum('VK', vk_dict.items())
VK_TO_NAME = {
    vk: name
    for name, vk in vk_dict.items()
}

VK_TO_KEYNAME = {
    VK.ADD                : 'kp_add',
    VK.APPS               : 'menu',
    VK.BACK               : 'backspace',
    VK.BROWSER_BACK       : 'back',
    VK.BROWSER_FAVORITES  : 'favorites',
    VK.BROWSER_FORWARD    : 'forward',
    VK.BROWSER_HOME       : 'homepage',
    VK.BROWSER_HOME       : 'www',
    VK.BROWSER_REFRESH    : 'refresh',
    VK.BROWSER_SEARCH     : 'search',
    VK.BROWSER_STOP       : 'stop',
    VK.CANCEL             : 'cancel',
    VK.CAPITAL            : 'caps_lock',
    VK.CLEAR              : 'clear',
    VK.DECIMAL            : 'kp_decimal',
    VK.DELETE             : 'delete',
    VK.DIVIDE             : 'kp_divide',
    VK.DOWN               : 'down',
    VK.END                : 'end',
    VK.ESCAPE             : 'escape',
    VK.EXECUTE            : 'execute',
    VK.F1                 : 'f1',
    VK.F2                 : 'f2',
    VK.F3                 : 'f3',
    VK.F4                 : 'f4',
    VK.F5                 : 'f5',
    VK.F6                 : 'f6',
    VK.F7                 : 'f7',
    VK.F8                 : 'f8',
    VK.F9                 : 'f9',
    VK.F10                : 'f10',
    VK.F11                : 'f11',
    VK.F12                : 'f12',
    VK.F13                : 'f13',
    VK.F14                : 'f14',
    VK.F15                : 'f15',
    VK.F16                : 'f16',
    VK.F17                : 'f17',
    VK.F18                : 'f18',
    VK.F19                : 'f19',
    VK.F20                : 'f20',
    VK.F21                : 'f21',
    VK.F22                : 'f22',
    VK.F23                : 'f23',
    VK.F24                : 'f24',
    VK.HELP               : 'help',
    VK.HOME               : 'home',
    VK.INSERT             : 'insert',
    VK.LAUNCH_APP1        : 'mycomputer',
    VK.LAUNCH_APP2        : 'calculator',
    VK.LAUNCH_MAIL        : 'mail',
    VK.LAUNCH_MEDIA_SELECT: 'audiomedia',
    VK.LCONTROL           : 'control_l',
    VK.LEFT               : 'left',
    VK.LMENU              : 'alt_l',
    VK.LSHIFT             : 'shift_l',
    VK.LWIN               : 'super_l',
    VK.MEDIA_NEXT_TRACK   : 'audionext',
    VK.MEDIA_PLAY_PAUSE   : 'audiopause',
    VK.MEDIA_PLAY_PAUSE   : 'audioplay',
    VK.MEDIA_PREV_TRACK   : 'audioprev',
    VK.MEDIA_STOP         : 'audiostop',
    VK.MODECHANGE         : 'mode_switch',
    VK.MULTIPLY           : 'kp_multiply',
    VK.NEXT               : 'page_down',
    VK.NUMLOCK            : 'num_lock',
    VK.NUMPAD0            : 'kp_0',
    VK.NUMPAD1            : 'kp_1',
    VK.NUMPAD2            : 'kp_2',
    VK.NUMPAD3            : 'kp_3',
    VK.NUMPAD4            : 'kp_4',
    VK.NUMPAD5            : 'kp_5',
    VK.NUMPAD6            : 'kp_6',
    VK.NUMPAD7            : 'kp_7',
    VK.NUMPAD8            : 'kp_8',
    VK.NUMPAD9            : 'kp_9',
    VK.OEM_PLUS           : 'kp_equal',
    VK.PAUSE              : 'pause',
    VK.PRIOR              : 'page_up',
    VK.RCONTROL           : 'control_r',
    VK.RETURN             : 'return',
    VK.RIGHT              : 'right',
    VK.RMENU              : 'alt_r',
    VK.RSHIFT             : 'shift_r',
    VK.RWIN               : 'super_r',
    VK.SCROLL             : 'scroll_lock',
    VK.SELECT             : 'select',
    VK.SLEEP              : 'standby',
    VK.SNAPSHOT           : 'print',
    VK.SUBTRACT           : 'kp_subtract',
    VK.TAB                : 'tab',
    VK.UP                 : 'up',
    VK.VOLUME_DOWN        : 'audiolowervolume',
    VK.VOLUME_MUTE        : 'audiomute',
    VK.VOLUME_UP          : 'audioraisevolume',
}

def vk_to_str(vk):
    s = VK_TO_NAME.get(vk)
    return '%x' % vk if s is None else s

# }}}


DEAD_KEYNAME = {
    'asciicircum': 'dead_circumflex',
    'asciitilde' : 'dead_tilde',
}


class KeyboardLayout: # {{{

    def __init__(self, layout_id=None, debug=False):
        if layout_id is None:
            layout_id = KeyboardLayout.current_layout_id()
        self.layout_id = layout_id

        # Find virtual key code for each scan code (if any).
        sc_to_vk = {}
        vk_to_sc = {}
        for sc in range(0x01, 0x7f + 1):
            vk = MapVirtualKeyEx(sc, 3, layout_id)
            if vk != 0:
                sc_to_vk[sc] = vk
                vk_to_sc[vk] = sc

        state = (wintypes.BYTE * 256)()
        strbuf = ctypes.create_unicode_buffer(8)

        def fill_state(ss):
            for mod_state, mod_vk in (
                (SHIFT_STATE.SHIFT, VK.SHIFT),
                (SHIFT_STATE.CTRL,  VK.CONTROL),
                (SHIFT_STATE.MENU,  VK.MENU),
            ):
                state[mod_vk] = 0x80 if (ss & mod_state) != 0 else 0

        def to_unichr(vk, sc, ss):
            fill_state(ss)
            rc = ToUnicodeEx(vk, sc,
                             state, strbuf, len(strbuf),
                             0, layout_id)
            if rc > 0:
                dead_key = False
                char = ctypes.wstring_at(strbuf.value, rc)
            elif rc < 0:
                # Dead key, flush state.
                dead_key = True
                fill_state(0)
                rc = ToUnicodeEx(VK.SPACE, vk_to_sc[VK.SPACE],
                                 state, strbuf, len(strbuf),
                                 0, layout_id)
                char = ctypes.wstring_at(strbuf.value, rc) if rc > 0 else ''
            else:
                dead_key = False
                char = ''
            return char, dead_key

        def sort_vk_ss_list(vk_ss_list):
            # Prefer lower modifiers combo.
            return sorted(vk_ss_list, key=lambda vk_ss: popcount_8(vk_ss[1]))

        char_to_vk_ss = defaultdict(list)
        keyname_to_vk = defaultdict(list)

        for sc, vk in sorted(sc_to_vk.items()):
            for ss in SHIFT_STATE:
                if ss in (SHIFT_STATE.MENU, SHIFT_STATE.SHIFT_MENU):
                    # Alt and Shift+Alt don't work, so skip them.
                    continue
                char, dead_key = to_unichr(vk, sc, ss)
                if debug and char:
                    print('%s%s -> %s [%r] %s' % (
                        shift_state_str(ss), vk_to_str(vk),
                        char, char, '[dead key]' if dead_key else '',
                    ))
                kn = None
                if char:
                    kn = CHAR_TO_KEYNAME.get(char)
                    if dead_key:
                        kn = kn and DEAD_KEYNAME.get(kn, 'dead_' + kn)
                    else:
                        char_to_vk_ss[char].append((vk, ss))
                    if kn:
                        keyname_to_vk[kn].append((vk, ss))

        self.char_to_vk_ss = {
            char: sort_vk_ss_list(vk_ss_list)[0]
            for char, vk_ss_list in char_to_vk_ss.items()
        }
        self.char_to_vk_ss['\n'] = self.char_to_vk_ss['\r']

        self.keyname_to_vk = {
            kn: vk
            for vk, kn in VK_TO_KEYNAME.items()
        }
        self.keyname_to_vk.update({
            'next'      : VK.NEXT,
            'prior'     : VK.PRIOR,
            'kp_delete' : VK.DELETE,
            'kp_enter'  : VK.RETURN,
            'break'     : VK.CANCEL,
        })
        self.keyname_to_vk.update({
            kn: sort_vk_ss_list(vk_ss_list)[0][0]
            for kn, vk_ss_list in keyname_to_vk.items()
        })
        add_modifiers_aliases(self.keyname_to_vk)

        self.ss_to_vks = {}
        for ss in SHIFT_STATE:
            vk_list = []
            for mod_state, mod_vk in (
                (SHIFT_STATE.SHIFT, VK.SHIFT),
                (SHIFT_STATE.CTRL,  VK.CONTROL),
                (SHIFT_STATE.MENU,  VK.MENU),
            ):
                if (ss & mod_state) != 0:
                    vk_list.append(mod_vk)
            self.ss_to_vks[ss] = tuple(vk_list)

    @staticmethod
    def current_layout_id():
        pid = GetWindowThreadProcessId(windll.user32.GetForegroundWindow(), 0)
        return GetKeyboardLayout(pid)

# }}}


if __name__ == '__main__':
    sys.stdout = codecs.getwriter('utf8')(sys.stdout)
    layout = KeyboardLayout(debug=True)
    print('character to virtual key + shift state [%u]' % len(layout.char_to_vk_ss))
    for char, combo in sorted(layout.char_to_vk_ss.items()):
        vk, ss = combo
        print('%s [%r:%s] -> %s%s' % (char, char,
                                      CHAR_TO_KEYNAME.get(char, '?'),
                                      shift_state_str(ss), vk_to_str(vk)))
    print()
    print('keyname to virtual key [%u]' % len(layout.keyname_to_vk))
    for kn, vk in sorted(layout.keyname_to_vk.items()):
        print('%s -> %s' % (kn, vk_to_str(vk)))
    print()
    print('modifiers combo [%u]' % len(layout.ss_to_vks))
    for ss, vk_list in sorted(layout.ss_to_vks.items()):
        print('%s -> %s' % (shift_state_str(ss), '+'.join(vk_to_str(vk) for vk in vk_list)))

# vim: foldmethod=marker



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

import ast

import pytest

from plover.dictionary.rtfcre_dict import RtfDictionary, format_translation
from plover.steno import normalize_steno

from . import parametrize
from .utils import make_dict


RTF_LOAD_TESTS = (

    # Empty dictionary.
    lambda: '''



    ''',

    # Only one translation.
    lambda: r'''
    {\*\cxs SP}translation

    'SP': 'translation',
    ''',

    # One translation on multiple lines.
    lambda: pytest.mark.xfail('''
    {\\*\\cxs SP}\r\ntranslation

    'SP': 'translation'
    '''),

    # Multiple translations no newlines.
    lambda: r'''
    {\*\cxs SP}translation{\*\cxs S}translation2

    'SP': 'translation',
    'S': 'translation2',
    ''',

    # Multiple translations on separate lines.
    lambda: '''
    {\\*\\cxs SP}translation\r\n{\\*\\cxs S}translation2

    'SP': 'translation',
    'S': 'translation2',
    ''',
    lambda: '''
    {\\*\\cxs SP}translation\n{\\*\\cxs S}translation2

    'SP': 'translation',
    'S': 'translation2',
    ''',

    # Escaped \r and \n handled
    lambda: '''
    {\\*\\cxs SP}trans\\\r\\\n

    'SP': 'trans{#Return}{#Return}{#Return}{#Return}',
    ''',

    # Escaped \r\n handled in mid translation
    lambda: '''
    {\\*\\cxs SP}trans\\\r\\\nlation

    'SP': 'trans{#Return}{#Return}{#Return}{#Return}lation',
    ''',

    # Whitespace is preserved in various situations.
    lambda: r'''
    {\*\cxs S}t  

    'S': 't{^  ^}',
    ''',
    lambda: r'''
    {\*\cxs S}  t

    'S': '{^  ^}t',
    ''',
    lambda: r'''
    {\*\cxs S}t   {\*\cxs T}t    

    'S': 't{^   ^}',
    'T': 't{^    ^}',
    ''',
    lambda: '''
    {\\*\\cxs S}t   \r\n{\\*\\cxs T}t    

    'S': 't{^   ^}',
    'T': 't{^    ^}',
    ''',
    lambda: '''
    {\\*\\cxs S}t  \r\n{\\*\\cxs T} t \r\n

    'S': 't{^  ^}',
    'T': ' t ',
    ''',

    # Translations are ignored if converter returns None
    lambda: r'''
    {\*\cxs S}return_none

    'S': 'return_none',
    ''',

    lambda: r'''
    {\*\cxs T}t t t  

    'T': 't t t{^  ^}',
    ''',

    # Conflicts result on only the last one kept.
    lambda: r'''
    {\*\cxs T}t
    {\*\cxs T}g

    'T': 'g',
    ''',
    lambda: r'''
    {\*\cxs T}t
    {\*\cxs T}return_none

    'T': 'return_none',
    ''',

    # Translation conversion tests.

    lambda: ('', ''),
    lambda: (r'\-', '-'),
    lambda: (r'\\ ', '\\ '),
    lambda: pytest.mark.xfail((r'\\', '\\')),
    lambda: (r'\{', '{'),
    lambda: (r'\}', '}'),
    lambda: (r'\~', '{^ ^}'),
    lambda: (r'\_', '-'),
    lambda: ('\\\r', '{#Return}{#Return}'),
    lambda: ('\\\n', '{#Return}{#Return}'),
    lambda: (r'\cxds', '{^}'),
    lambda: (r'pre\cxds ', '{pre^}'),
    lambda: (r'pre\cxds  ', '{pre^} '),
    lambda: (r'pre\cxds', '{pre^}'),
    lambda: (r'\cxds post', '{^post}'),
    lambda: (r'\cxds in\cxds', '{^in^}'),
    lambda: (r'\cxds in\cxds ', '{^in^}'),
    lambda: (r'\cxfc', '{-|}'),
    lambda: (r'\cxfl', '{>}'),
    lambda: (r'pre\cxfl', 'pre{>}'),
    lambda: (r'\par', '{#Return}{#Return}'),
    # Stenovations extensions...
    lambda: (r'{\*\cxsvatdictflags N}', '{-|}'),
    lambda: (r'{\*\cxsvatdictflags LN1}', '{-|}'),
    # caseCATalyst declares new styles without a preceding \par so we treat
    # it as an implicit par.
    lambda: (r'\s1', '{#Return}{#Return}'),
    # But if the \par is present we don't treat \s as an implicit par.
    lambda: (r'\par\s1', '{#Return}{#Return}'),
    # Continuation styles are indented too.
    lambda: (r'\par\s4', '{#Return}{#Return}{^    ^}'),
    # caseCATalyst punctuation.
    lambda: (r'.', '{.}'),
    lambda: (r'. ', '{.} '),
    lambda: (r' . ', ' . '),
    lambda: (r'{\cxa Q.}.', 'Q..'),
    # Don't mess with period that is part of a word.
    lambda: (r'Mr.', 'Mr.'),
    lambda: (r'.attribute', '.attribute'),
    lambda: (r'{\cxstit contents}', 'contents'),
    lambda: (r'{\cxfing c}', '{&c}'),
    lambda: (r'{\cxp.}', '{.}'),
    lambda: (r'{\cxp .}', '{.}'),
    lambda: (r'{\cxp . }', '{.}'),
    lambda: (r'{\cxp .  }', '{.}'),
    lambda: (r'{\cxp !}', '{!}'),
    lambda: (r'{\cxp ?}', '{?}'),
    lambda: (r'{\cxp ,}', '{,}'),
    lambda: (r'{\cxp ;}', '{;}'),
    lambda: (r'{\cxp :}', '{:}'),
    lambda: ('{\\cxp \'}', '{^\'}'),
    lambda: ('{\\cxp -}', '{^-^}'),
    lambda: ('{\\cxp /}', '{^/^}'),
    lambda: ('{\\cxp...  }', '{^...  ^}'),
    lambda: ('{\\cxp ") }', '{^") ^}'),
    lambda: ('{\\nonexistent }', ''),
    lambda: ('{\\nonexistent contents}', 'contents'),
    lambda: ('{\\nonexistent cont\\_ents}', 'cont-ents'),
    lambda: ('{\\*\\nonexistent }', ''),
    lambda: ('{\\*\\nonexistent contents}', ''),
    lambda: ('{eclipse command}', '{eclipse command}'),
    lambda: ('test text', 'test text'),
    lambda: ('test  text', 'test{^  ^}text'),
    lambda: (r'{\cxconf [{\cxc abc}]}', 'abc'),
    lambda: (r'{\cxconf [{\cxc abc}|{\cxc def}]}', 'def'),
    lambda: (r'{\cxconf [{\cxc abc}|{\cxc def}|{\cxc ghi}]}', 'ghi'),
    lambda: (r'{\cxconf [{\cxc abc}|{\cxc {\cxp... }}]}', '{^... ^}'),
    lambda: (r'be\cxds{\*\cxsvatdictentrydate\yr2006\mo5\dy10}', '{be^}'),
    lambda: (r'{\nonexistent {\cxp .}}', '{.}'),
    lambda: (r'{\*\nonexistent {\cxp .}}', ''),
)

@parametrize(RTF_LOAD_TESTS, arity=1)
def test_rtf_load(test_case):
    if isinstance(test_case, tuple):
        # Translation conversion test.
        rtf_entries = r'{\*\cxs S}' + test_case[0]
        dict_entries = { normalize_steno('S'): test_case[1] }
    else:
        rtf_entries, dict_entries = test_case.rsplit('\n\n', 1)
        dict_entries = {
            normalize_steno(k): v
            for k, v in ast.literal_eval('{' + dict_entries + '}').items()
        }
    rtf_styles = {
        0: 'Normal',
        1: 'Question',
        2: 'Answer',
        3: 'Colloquy',
        4: 'Continuation Q',
        5: 'Continuation A',
        6: 'Continuation Col',
        7: 'Paren',
        8: 'Centered',
    }
    rtf = (
        '\r\n'.join(
            [r'{\rtf1\ansi\cxdict{\*\cxrev100}{\*\cxsystem Fake Software}'] +
            [r'{\s%d %s;}' % (k, v) for k, v in rtf_styles.items()] +
            ['}'])
        + rtf_entries
        + '\r\n}'
    )
    with make_dict(rtf.encode('cp1252')) as filename:
        d = RtfDictionary.load(filename)
        assert dict(d.items()) == dict_entries



@parametrize((
    lambda: ('', ''),
    lambda: ('{^in^}', r'\cxds in\cxds '),
    lambda: ('{pre^}', r'pre\cxds '),
    lambda: ('{pre^} ', r'pre\cxds '),
    lambda: ('{pre^}  ', r'pre\cxds '),
))
def test_format_translation(before, expected):
    result = format_translation(before)
    assert result == expected


@parametrize((
    lambda: ({'S/T': '{pre^}'},
     b'{\\rtf1\\ansi{\\*\\cxrev100}\\cxdict{\\*\\cxsystem Plover}'
     b'{\\stylesheet{\\s0 Normal;}}\r\n'
     b'{\\*\\cxs S///T}pre\\cxds \r\n}\r\n'
    ),
))
def test_save_dictionary(contents, expected):
    with make_dict(b'foo') as filename:
        d = RtfDictionary.create(filename)
        d.update(contents)
        d.save()
        with open(filename, 'rb') as fp:
            contents = fp.read()
    assert contents == expected




from io import open
import os
import re
import collections

from plover.oslayer.config import CONFIG_DIR, ASSETS_DIR
from plover.registry import registry


def _load_wordlist(filename):
    if filename is None:
        return {}
    path = None
    for dir in (CONFIG_DIR, ASSETS_DIR):
        path = os.path.realpath(os.path.join(dir, filename))
        if os.path.exists(path):
            break
    words = {}
    with open(path, encoding='utf-8') as f:
        pairs = [word.strip().rsplit(' ', 1) for word in f]
        pairs.sort(reverse=True, key=lambda x: int(x[1]))
        words = {p[0]: int(p[1]) for p in pairs}
    return words

def _key_order(keys, numbers):
    key_order = {}
    for order, key in enumerate(keys):
        key_order[key] = order
        number_key = numbers.get(key)
        if number_key is not None:
            key_order[number_key] = order
    return key_order

def _suffix_keys(keys):
    assert isinstance(keys, collections.Sequence)
    return keys

_EXPORTS = {
    'KEYS'                     : lambda mod: mod.KEYS,
    'KEY_ORDER'                : lambda mod: _key_order(mod.KEYS, mod.NUMBERS),
    'NUMBER_KEY'               : lambda mod: mod.NUMBER_KEY,
    'NUMBERS'                  : lambda mod: dict(mod.NUMBERS),
    'SUFFIX_KEYS'              : lambda mod: _suffix_keys(mod.SUFFIX_KEYS),
    'UNDO_STROKE_STENO'        : lambda mod: mod.UNDO_STROKE_STENO,
    'IMPLICIT_HYPHEN_KEYS'     : lambda mod: set(mod.IMPLICIT_HYPHEN_KEYS),
    'IMPLICIT_HYPHENS'         : lambda mod: {l.replace('-', '')
                                              for l in mod.IMPLICIT_HYPHEN_KEYS},
    'ORTHOGRAPHY_WORDS'        : lambda mod: _load_wordlist(mod.ORTHOGRAPHY_WORDLIST),
    'ORTHOGRAPHY_RULES'        : lambda mod: [(re.compile(pattern, re.I), replacement)
                                              for pattern, replacement in mod.ORTHOGRAPHY_RULES],
    'ORTHOGRAPHY_RULES_ALIASES': lambda mod: dict(mod.ORTHOGRAPHY_RULES_ALIASES),
    'KEYMAPS'                  : lambda mod: mod.KEYMAPS,
    'DICTIONARIES_ROOT'        : lambda mod: mod.DICTIONARIES_ROOT,
    'DEFAULT_DICTIONARIES'     : lambda mod: mod.DEFAULT_DICTIONARIES,
}

def setup(system_name):
    system_symbols = {}
    mod = registry.get_plugin('system', system_name).obj
    for symbol, init in _EXPORTS.items():
        system_symbols[symbol] = init(mod)
    system_symbols['NAME'] = system_name
    globals().update(system_symbols)

NAME = None



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

from plover.orthography import add_suffix

from . import parametrize


ADD_SUFFIX_TESTS = (
    lambda: ('artistic', 'ly', 'artistically'),
    lambda: ('cosmetic', 'ly', 'cosmetically'),
    lambda: ('establish', 's', 'establishes'),
    lambda: ('speech', 's', 'speeches'),
    lambda: ('approach', 's', 'approaches'),
    lambda: ('beach', 's', 'beaches'),
    lambda: ('arch', 's', 'arches'),
    lambda: ('larch', 's', 'larches'),
    lambda: ('march', 's', 'marches'),
    lambda: ('search', 's', 'searches'),
    lambda: ('starch', 's', 'starches'),
    lambda: ('stomach', 's', 'stomachs'),
    lambda: ('monarch', 's', 'monarchs'),
    lambda: ('patriarch', 's', 'patriarchs'),
    lambda: ('oligarch', 's', 'oligarchs'),
    lambda: ('cherry', 's', 'cherries'),
    lambda: ('day', 's', 'days'),
    lambda: ('penny', 's', 'pennies'),
    lambda: ('pharmacy', 'ist', 'pharmacist'),
    lambda: ('melody', 'ist', 'melodist'),
    lambda: ('pacify', 'ist', 'pacifist'),
    lambda: ('geology', 'ist', 'geologist'),
    lambda: ('metallurgy', 'ist', 'metallurgist'),
    lambda: ('anarchy', 'ist', 'anarchist'),
    lambda: ('monopoly', 'ist', 'monopolist'),
    lambda: ('alchemy', 'ist', 'alchemist'),
    lambda: ('botany', 'ist', 'botanist'),
    lambda: ('therapy', 'ist', 'therapist'),
    lambda: ('theory', 'ist', 'theorist'),
    lambda: ('psychiatry', 'ist', 'psychiatrist'),
    lambda: ('lobby', 'ist', 'lobbyist'),
    lambda: ('hobby', 'ist', 'hobbyist'),
    lambda: ('copy', 'ist', 'copyist'),
    lambda: ('beauty', 'ful', 'beautiful'),
    lambda: ('weary', 'ness', 'weariness'),
    lambda: ('weary', 'some', 'wearisome'),
    lambda: ('lonely', 'ness', 'loneliness'),
    lambda: ('narrate', 'ing', 'narrating'),
    lambda: ('narrate', 'or', 'narrator'),
    lambda: ('generalize', 'ability', 'generalizability'),
    lambda: ('reproduce', 'able', 'reproducible'),
    lambda: ('grade', 'ations', 'gradations'),
    lambda: ('urine', 'ary', 'urinary'),
    lambda: ('achieve', 'able', 'achievable'),
    lambda: ('polarize', 'ation', 'polarization'),
    lambda: ('done', 'or', 'donor'),
    lambda: ('analyze', 'ed', 'analyzed'),
    lambda: ('narrate', 'ing', 'narrating'),
    lambda: ('believe', 'able', 'believable'),
    lambda: ('animate', 'ors', 'animators'),
    lambda: ('discontinue', 'ation', 'discontinuation'),
    lambda: ('innovate', 'ive', 'innovative'),
    lambda: ('future', 'ists', 'futurists'),
    lambda: ('illustrate', 'or', 'illustrator'),
    lambda: ('emerge', 'ent', 'emergent'),
    lambda: ('equip', 'ed', 'equipped'),
    lambda: ('defer', 'ed', 'deferred'),
    lambda: ('defer', 'er', 'deferrer'),
    lambda: ('defer', 'ing', 'deferring'),
    lambda: ('pigment', 'ed', 'pigmented'),
    lambda: ('refer', 'ed', 'referred'),
    lambda: ('fix', 'ed', 'fixed'),
    lambda: ('alter', 'ed', 'altered'),
    lambda: ('interpret', 'ing', 'interpreting'),
    lambda: ('wonder', 'ing', 'wondering'),
    lambda: ('target', 'ing', 'targeting'),
    lambda: ('limit', 'er', 'limiter'),
    lambda: ('maneuver', 'ing', 'maneuvering'),
    lambda: ('monitor', 'ing', 'monitoring'),
    lambda: ('color', 'ing', 'coloring'),
    lambda: ('inhibit', 'ing', 'inhibiting'),
    lambda: ('master', 'ed', 'mastered'),
    lambda: ('target', 'ing', 'targeting'),
    lambda: ('fix', 'ed', 'fixed'),
    lambda: ('scrap', 'y', 'scrappy'),
    lambda: ('trip', 's', 'trips'),
    lambda: ('equip', 's', 'equips'),
    lambda: ('bat', 'en', 'batten'),
    lambda: ('smite', 'en', 'smitten'),
    lambda: ('got', 'en', 'gotten'),
    lambda: ('bite', 'en', 'bitten'),
    lambda: ('write', 'en', 'written'),
    lambda: ('flax', 'en', 'flaxen'),
    lambda: ('wax', 'en', 'waxen'),
    lambda: ('fast', 'est', 'fastest'),
    lambda: ('white', 'er', 'whiter'),
    lambda: ('crap', 'y', 'crappy'),
    lambda: ('lad', 'er', 'ladder'),
    lambda: ('translucent', 'cy', 'translucency'),
    lambda: ('bankrupt', 'cy', 'bankruptcy'),
    lambda: ('inadequate', 'cy', 'inadequacy'),
    lambda: ('secret', 'cy', 'secrecy'),
    lambda: ('impolite', 'cy', 'impolicy'),
    lambda: ('idiot', 'cy', 'idiocy'),
    lambda: ('free', 'ed', 'freed'),
    lambda: ('free', 'er', 'freer'),
    lambda: ('regulate', 'ry', 'regulatory'),
)

@parametrize(ADD_SUFFIX_TESTS)
def test_add_suffix(word, suffix, expected):
    assert add_suffix(word, suffix) == expected



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.
#
# TODO: Convert non-ascii characters to UTF8
# TODO: What does ^ mean in Eclipse?
# TODO: What does #N mean in Eclipse?
# TODO: convert supported commands from Eclipse

"""Parsing an RTF/CRE dictionary.

RTF/CRE spec:
http://www.legalxml.org/workgroups/substantive/transcripts/cre-spec.htm

"""

import codecs
import inspect
import re

from plover.steno import normalize_steno
from plover.steno_dictionary import StenoDictionary
# TODO: Move dictionary format somewhere more canonical than formatting.
from plover.formatting import META_RE


# A regular expression to capture an individual entry in the dictionary.
DICT_ENTRY_PATTERN = re.compile(r'(?s)(?<!\\){\\\*\\cxs (?P<steno>[^}]+)}' + 
                                r'(?P<translation>.*?)(?:(?<!\\)(?:\r\n|\n))*?'+
                                r'(?=(?:(?<!\\){\\\*\\cxs [^}]+})|' +
                                r'(?:(?:(?<!\\)(?:\r\n|\n)\s*)*}\s*\Z))')

class TranslationConverter:
    """Convert an RTF/CRE translation into plover's internal format."""
    
    def __init__(self, styles={}):
        self.styles = styles
        
        def linenumber(f):
            return f[1].__code__.co_firstlineno
        
        handler_funcs = inspect.getmembers(self, inspect.ismethod)
        handler_funcs.sort(key=linenumber)
        handlers = [self._make_re_handler(f.__doc__, f)
                    for name, f in handler_funcs 
                    if name.startswith('_re_handle_')]
        handlers.append(self._match_nested_command_group)
        def handler(s, pos):
            for handler in handlers:
                result = handler(s, pos)
                if result:
                    return result
            return None
        self._handler = handler
        self._command_pattern = re.compile(
            r'(\\\*)?\\([a-z]+)(-?[0-9]+)?[ ]?')
        self._multiple_whitespace_pattern = re.compile(r'([ ]{2,})')
        # This poorly named variable indicates whether the current context is
        # one where commands can be inserted (True) or not (False).
        self._whitespace = True
    
    def _make_re_handler(self, pattern, f):
        pattern = re.compile(pattern)
        def handler(s, pos):
            match = pattern.match(s, pos)
            if match:
                newpos = match.end()
                result = f(match)
                return (newpos, result)
            return None
        return handler

    def _re_handle_escapedchar(self, m):
        r'\\([-\\{}])'
        return m.group(1)
        
    def _re_handle_hardspace(self, m):
        r'\\~'
        return '{^ ^}'
        
    def _re_handle_dash(self, m):
        r'\\_'
        return '-'
        
    def _re_handle_escaped_newline(self, m):
        r'\\\r|\\\n'
        return '{#Return}{#Return}'
        
    def _re_handle_infix(self, m):
        r'\\cxds ([^{}\\\r\n]+)\\cxds ?'
        return '{^%s^}' % m.group(1)
        
    def _re_handle_suffix(self, m):
        r'\\cxds ([^{}\\\r\n ]+)'
        return '{^%s}' % m.group(1)

    def _re_handle_prefix(self, m):
        r'([^{}\\\r\n ]+)\\cxds ?'
        return '{%s^}' % m.group(1)

    def _re_handle_commands(self, m):
        r'(\\\*)?\\([a-z]+)(-?[0-9]+)? ?'
        
        command = m.group(2)
        arg = m.group(3)
        if arg:
            arg = int(arg)
        
        if command == 'cxds':
            return '{^}'
        
        if command == 'cxfc':
            return '{-|}'

        if command == 'cxfl':
            return '{>}'

        if command == 'par':
            self.seen_par = True
            return '{#Return}{#Return}'
            
        if command == 's':
            result = []
            if not self.seen_par:
                result.append('{#Return}{#Return}')
            style_name = self.styles.get(arg, '')
            if style_name.startswith('Contin'):
                result.append('{^    ^}')
            return ''.join(result)

        # Unrecognized commands are ignored.
        return ''

    def _re_handle_simple_command_group(self, m):
        r'{(\\\*)?\\([a-z]+)(-?[0-9]+)?[ ]?([^{}]*)}'
        
        ignore = bool(m.group(1))
        command = m.group(2)
        contents = m.group(4)
        if contents is None:
            contents = ''

        if command == 'cxstit':
            # Plover doesn't support stitching.
            return self(contents)
        
        if command == 'cxfing':
            prev = self._whitespace
            self._whitespace = False
            result = '{&' + contents + '}'
            self._whitespace = prev
            return result
            
        if command == 'cxp':
            prev = self._whitespace
            self._whitespace = False
            contents = self(contents)
            if contents is None:
                return None
            self._whitespace = prev
            stripped = contents.strip()
            if stripped in ['.', '!', '?', ',', ';', ':']:
                return '{' + stripped + '}'
            if stripped == "'":
                return "{^'}"
            if stripped in ['-', '/']:
                return '{^' + contents + '^}'
            # Show unknown punctuation as given.
            return '{^' + contents + '^}'
        
        if command == 'cxsvatdictflags' and 'N' in contents:
            return '{-|}'
        
        # unrecognized commands
        if ignore:
            return ''
        else:
            return self(contents)

    def _re_handle_eclipse_command(self, m):
        r'({[^\\][^{}]*})'
        return m.group()

    # caseCATalyst doesn't put punctuation in \cxp so we will treat any 
    # isolated punctuation at the beginning of the translation as special.
    def _re_handle_punctuation(self, m):
        r'^([.?!:;,])(?=\s|$)'
        if self._whitespace:
            result = '{%s}' % m.group(1)
        else:
            result = m.group(1)
        return result

    def _re_handle_text(self, m):
        r'[^{}\\\r\n]+'
        text = m.group()
        if self._whitespace:
            text = self._multiple_whitespace_pattern.sub(r'{^\1^}', text)
        return text

    def _get_matching_bracket(self, s, pos):
        if s[pos] != '{':
            return None
        end = len(s)
        depth = 1
        pos += 1
        while pos != end:
            c = s[pos]
            if c == '{':
                depth += 1
            elif c == '}':
                depth -= 1
            if depth == 0:
                break
            pos += 1
        if pos < end and s[pos] == '}':
            return pos
        return None

    def _get_command(self, s, pos):
        return self._command_pattern.match(s, pos)

    def _match_nested_command_group(self, s, pos):
        startpos = pos
        endpos = self._get_matching_bracket(s, pos)
        if endpos is None:
            return None

        command_match = self._get_command(s, startpos + 1)
        if command_match is None:
            return None

        ignore = bool(command_match.group(1))
        command = command_match.group(2)
        
        if command == 'cxconf':
            pos = command_match.end()
            last = ''
            while pos < endpos:
                if s[pos] in ['[', '|', ']']:
                    pos += 1
                    continue
                if s[pos] == '{':
                    command_match = self._get_command(s, pos + 1)
                    if command_match is None:
                        return None
                    if command_match.group(2) != 'cxc':
                        return None
                    cxc_end = self._get_matching_bracket(s, pos)
                    if cxc_end is None:
                        return None
                    last = s[command_match.end():cxc_end]
                    pos = cxc_end + 1
                    continue
                return None
            return (endpos + 1, self(last))
            
        if ignore:
            return (endpos + 1, '')
        else:
            return (endpos + 1, self(s[command_match.end():endpos]))

    def __call__(self, s):
        self.seen_par = False
        
        pos = 0
        tokens = []
        handler = self._handler
        end = len(s)
        while pos != end:
            result = handler(s, pos)
            if result is None:
                return None
            pos = result[0]
            token = result[1]
            if token is None:
                return None
            tokens.append(token)
        return ''.join(tokens)

STYLESHEET_RE = re.compile(r'(?s){\\s([0-9]+).*?((?:\b\w+\b\s*)+);}')

def load_stylesheet(s):
    """Returns a dictionary mapping a number to a style name."""
    return {int(k): v for k, v in STYLESHEET_RE.findall(s)}

HEADER = ("{\\rtf1\\ansi{\\*\\cxrev100}\\cxdict{\\*\\cxsystem Plover}" +
          "{\\stylesheet{\\s0 Normal;}}\r\n")

def format_translation(t):
    t = ' '.join([x.strip() for x in META_RE.findall(t) if x.strip()])
    
    t = re.sub(r'{\.}', r'{\\cxp. }', t)
    t = re.sub(r'{!}', r'{\\cxp! }', t)
    t = re.sub(r'{\?}', r'{\\cxp? }', t)
    t = re.sub(r'{\,}', r'{\\cxp, }', t)
    t = re.sub(r'{:}', r'{\\cxp: }', t)
    t = re.sub(r'{;}', r'{\\cxp; }', t)
    t = re.sub(r'{\^}', r'\\cxds ', t)
    t = re.sub(r'{\^([^^}]*)}', r'\\cxds \1', t)
    t = re.sub(r'{([^^}]*)\^}', r'\1\\cxds ', t)
    t = re.sub(r'{\^([^^}]*)\^}', r'\\cxds \1\\cxds ', t)
    t = re.sub(r'{-\|}', r'\\cxfc ', t)
    t = re.sub(r'{>}', r'\\cxfls ', t)
    t = re.sub(r'{ }', r' ', t)
    t = re.sub(r'{&([^}]+)}', r'{\\cxfing \1}', t)
    t = re.sub(r'{#([^}]+)}', r'\\{#\1\\}', t)
    t = re.sub(r'{PLOVER:([a-zA-Z]+)}', r'\\{PLOVER:\1\\}', t)
    t = re.sub(r'\\"', r'"', t)

    return t


class RtfDictionary(StenoDictionary):

    def _load(self, filename):
        with open(filename, 'rb') as fp:
            s = fp.read().decode('cp1252')
        def parse():
            styles = load_stylesheet(s)
            converter = TranslationConverter(styles)
            for m in DICT_ENTRY_PATTERN.finditer(s):
                steno = normalize_steno(m.group('steno'))
                translation = m.group('translation')
                converted = converter(translation)
                if converted is not None:
                    yield steno, converted
        self.update(parse())

    def _save(self, filename):
        with open(filename, 'wb') as fp:
            writer = codecs.getwriter('cp1252')(fp)
            writer.write(HEADER)
            for s, t in self.items():
                s = '/'.join(s)
                t = format_translation(t)
                entry = "{\\*\\cxs %s}%s\r\n" % (s, t)
                writer.write(entry)
            writer.write("}\r\n")




import os
import sys
import subprocess

from plover.oslayer.config import CONFIG_DIR, PLUGINS_PLATFORM


def main():
    args = sys.argv[:]
    args[0:1] = [sys.executable, '-m', 'plover.main']
    if '--no-user-plugins' in args[3:]:
        args.remove('--no-user-plugins')
        args.insert(1, '-s')
    os.environ['PYTHONUSERBASE'] = os.path.join(CONFIG_DIR, 'plugins', PLUGINS_PLATFORM)
    if sys.platform.startswith('win32'):
        # Workaround https://bugs.python.org/issue19066
        subprocess.Popen(args, cwd=os.getcwd())
        os.exit(0)
    os.execv(args[0], args)


if __name__ == '__main__':
    main()
